#!/usr/bin/env python

"""
Test CIM types.
"""

from __future__ import absolute_import, print_function

from datetime import timedelta, datetime
import pytest

from pywbem import CIMType, CIMInt, CIMFloat, Uint8, Uint16, Uint32, Uint64, \
    Sint8, Sint16, Sint32, Sint64, Real32, Real64, CIMDateTime, MinutesFromUTC


#
# CIM integer data types
#


@pytest.fixture(params=[

    # Each list item is a tuple of:
    # - obj_type (type): CIM integer data type (e.g. Uint8).
    # - exp_cimtype (string): Expected value of cimtype class attribute for
    #   this CIM data type.
    # - exp_minvalue (integer): Expected value of minvalue class attribute for
    #   this CIM data type.
    # - exp_maxvalue (): Expected value of maxvalue class attribute for
    #   this CIM data type.

    # (obj_type, exp_cimtype, exp_minvalue, exp_maxvalue)
    (Uint8, 'uint8', 0, 2**8 - 1),
    (Uint16, 'uint16', 0, 2**16 - 1),
    (Uint32, 'uint32', 0, 2**32 - 1),
    (Uint64, 'uint64', 0, 2**64 - 1),
    (Sint8, 'sint8', -2**7, 2**7 - 1),
    (Sint16, 'sint16', -2**15, 2**15 - 1),
    (Sint32, 'sint32', -2**31, 2**31 - 1),
    (Sint64, 'sint64', -2**63, 2**63 - 1),
], scope='module')
def integer_tuple(request):
    return request.param


def test_integer_class_attrs_class(integer_tuple):
    """Test class attrs via class level"""
    obj_type, exp_cimtype, exp_minvalue, exp_maxvalue = integer_tuple
    assert obj_type.cimtype == exp_cimtype
    assert obj_type.minvalue == exp_minvalue
    assert obj_type.maxvalue == exp_maxvalue


def test_integer_class_attrs_inst(integer_tuple):
    """Test class attrs via instance level"""
    obj_type, exp_cimtype, exp_minvalue, exp_maxvalue = integer_tuple
    obj = obj_type(42)
    assert obj.cimtype == exp_cimtype
    assert obj.minvalue == exp_minvalue
    assert obj.maxvalue == exp_maxvalue


def test_integer_inheritance(integer_tuple):
    """Test inheritance"""
    obj_type = integer_tuple[0]
    obj = obj_type(42)
    assert isinstance(obj, obj_type)
    assert isinstance(obj, CIMType)
    assert isinstance(obj, CIMInt)
    assert not isinstance(obj, CIMFloat)


def test_integer_init_int(integer_tuple):
    """Test initialization from integer value"""
    obj_type = integer_tuple[0]
    obj = obj_type(42)
    assert obj == 42


def test_integer_init_str(integer_tuple):
    """Test initialization from string value"""
    obj_type = integer_tuple[0]
    obj = obj_type('42')
    assert obj == 42


def test_integer_init_str_base10(integer_tuple):
    """Test initialization from string value with base 10"""
    obj_type = integer_tuple[0]
    obj = obj_type('42', 10)
    assert obj == 42


def test_integer_init_str_base16(integer_tuple):
    """Test initialization from string value with base 16"""
    obj_type = integer_tuple[0]
    obj = obj_type('2A', 16)
    assert obj == 42


def test_integer_init_minimum(integer_tuple):
    """Test initialization from integer value at minimum"""
    obj_type = integer_tuple[0]
    exp_minvalue = integer_tuple[2]
    obj = obj_type(exp_minvalue)
    assert obj == exp_minvalue


def test_integer_init_maximum(integer_tuple):
    """Test initialization from integer value at maximum"""
    obj_type = integer_tuple[0]
    exp_maxvalue = integer_tuple[3]
    obj = obj_type(exp_maxvalue)
    assert obj == exp_maxvalue


def test_integer_init_too_low(integer_tuple):
    """Test initialization from integer value below minimum"""
    obj_type = integer_tuple[0]
    exp_minvalue = integer_tuple[2]
    try:
        obj_type(exp_minvalue - 1)
    except ValueError:
        pass
    else:
        raise AssertionError("ValueError was not raised.")


def test_integer_init_too_high(integer_tuple):
    """Test initialization from integer value above maximum"""
    obj_type = integer_tuple[0]
    exp_maxvalue = integer_tuple[3]
    try:
        obj_type(exp_maxvalue + 1)
    except ValueError:
        pass
    else:
        raise AssertionError("ValueError was not raised.")


#
# CIM real data types
#


@pytest.fixture(params=[

    # Each list item is a tuple of:
    # - obj_type (type): CIM real data type (e.g. Real32).
    # - exp_cimtype (string): Expected value of cimtype class attribute for
    #   this CIM data type.

    # (obj_type, exp_cimtype)
    (Real32, 'real32'),
    (Real64, 'real64'),
], scope='module')
def real_tuple(request):
    return request.param


def test_real_class_attrs_class(real_tuple):
    """Test class attrs via class level"""
    obj_type, exp_cimtype = real_tuple
    assert obj_type.cimtype == exp_cimtype


def test_real_class_attrs_inst(real_tuple):
    """Test class attrs via instance level"""
    obj_type, exp_cimtype = real_tuple
    obj = obj_type(42)
    assert obj.cimtype == exp_cimtype


def test_real_inheritance(real_tuple):
    """Test inheritance"""
    obj_type = real_tuple[0]
    obj = obj_type(42)
    assert isinstance(obj, obj_type)
    assert isinstance(obj, CIMType)
    assert isinstance(obj, CIMFloat)
    assert not isinstance(obj, CIMInt)


def test_real_init_float(real_tuple):
    """Test initialization from floating point value"""
    obj_type = real_tuple[0]
    obj = obj_type(42.0)
    assert obj == 42.0


def test_real_init_str(real_tuple):
    """Test initialization from string value"""
    obj_type = real_tuple[0]
    obj = obj_type('42.0')
    assert obj == 42.0


#
# CIM datetime data type
#


@pytest.fixture(params=[

    # Each list item is a tuple of:
    # (
    #   dtarg: Input argument for CIMDateTime().
    #   exp_kind: Expected kind of datetime: 'interval' or 'timestamp', or
    #             exception type if initialization failed.
    #   exp_datetime: Expected value for datetime property.
    #   exp_timedelta: Expected value for timedelta property.
    #   exp_minutesfromutc: Expected value for minutes_from_utc property.
    #   exp_str: Expected string value of the CIMDateTime object.
    # )
    (
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        0,
        '12345678224455.654321:000'
    ),
    (
        '12345678224455.654321:000',
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        0,
        '12345678224455.654321:000'
    ),
    (
        CIMDateTime('12345678224455.654321:000'),
        'interval',
        None,
        timedelta(days=12345678, hours=22, minutes=44, seconds=55,
                  microseconds=654321),
        0,
        '12345678224455.654321:000'
    ),
    (
        '12345678224455.654321:777',
        ValueError, None, None, None, None
    ),
    (
        '12345678224455,654321:000',
        ValueError, None, None, None, None
    ),
    (
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        None,
        120,
        '20140924193040.654321+120'
    ),
    (
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(0)),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(0)),
        None,
        0,
        '20140924193040.654321+000'
    ),
    (
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(0)),
        None,
        0,
        '20140924193040.654321+000'
    ),
    (
        '20140924193040.654321+120',
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        None,
        120,
        '20140924193040.654321+120'
    ),
    (
        CIMDateTime('20140924193040.654321+120'),
        'timestamp',
        datetime(year=2014, month=9, day=24, hour=19, minute=30, second=40,
                 microsecond=654321, tzinfo=MinutesFromUTC(120)),
        None,
        120,
        '20140924193040.654321+120'
    ),
    (
        '20140924193040.654321*120',
        ValueError, None, None, None, None
    ),
    (
        '20140924193040,654321+120',
        ValueError, None, None, None, None
    ),
    (
        '20141324193040.654321+120',
        ValueError, None, None, None, None
    ),
    (
        '20140932193040.654321+120',
        ValueError, None, None, None, None
    ),
    (
        '20140924253040.654321+120',
        ValueError, None, None, None, None
    ),
    (
        '20140924196140.654321+120',
        ValueError, None, None, None, None
    ),
    (
        '20140924193061.654321+120',
        ValueError, None, None, None, None
    ),
    (
        42,
        TypeError, None, None, None, None
    ),
], scope='module')
def datetime_init_tuple(request):
    return request.param


def test_datetime_class_attrs_class():
    """Test class attrs via class level"""
    assert CIMDateTime.cimtype == 'datetime'


def test_datetime_class_attrs_inst():
    """Test class attrs via instance level"""
    obj = CIMDateTime('00000000000000.000000:000')
    assert obj.cimtype == 'datetime'


def test_datetime_inheritance():
    """Test inheritance"""
    obj = CIMDateTime('00000000000000.000000:000')
    assert isinstance(obj, CIMDateTime)
    assert isinstance(obj, CIMType)
    assert not isinstance(obj, CIMFloat)
    assert not isinstance(obj, CIMInt)


def test_datetime_init(datetime_init_tuple):
    """Test initialization from all input types"""
    (dtarg, exp_kind, exp_datetime, exp_timedelta, exp_minutesfromutc,
     exp_str) = datetime_init_tuple
    try:
        obj = CIMDateTime(dtarg)
    except Exception as exc:
        assert isinstance(exc, exp_kind)
    else:
        assert obj.is_interval == (exp_kind == 'interval')
        assert obj.datetime == exp_datetime
        if obj.datetime is not None:
            assert isinstance(obj.datetime, datetime)
            # We ensure that the datetime is always timezone-aware:
            assert obj.datetime.tzinfo is not None
        assert obj.timedelta == exp_timedelta
        if obj.timedelta is not None:
            assert isinstance(obj.timedelta, timedelta)
        assert obj.minutes_from_utc == exp_minutesfromutc
        assert str(obj) == exp_str


# TODO: Add testcases for get_local_utcoffset()
# TODO: Add testcases for now()
# TODO: Add testcases for fromtimestamp()



#!/usr/bin/env python
#
# Test ValueMapping class
#

from __future__ import absolute_import

import re
import unittest
from mock import Mock

from pywbem import CIMClass, CIMProperty, CIMQualifier, Uint8, WBEMServer, \
    WBEMConnection, ValueMapping

CLASSNAME = 'C1'
NAMESPACE = 'ns'
PROPNAME = 'p1'


class TestAll(unittest.TestCase):

    def setUp(self):
        self.conn = WBEMConnection('dummy')
        self.server = WBEMServer(self.conn)

    def setup_for_property(self, valuemap, values):
        """
        Set up a test class with a ValueMap/Values qualified property,
        and mock the GetClass() method to return that test class.
        """
        test_prop = CIMProperty(PROPNAME, None, 'uint8')
        test_prop.qualifiers['ValueMap'] = \
            CIMQualifier('ValueMap', valuemap, 'string')
        test_prop.qualifiers['Values'] = \
            CIMQualifier('Values', values, 'string')
        test_class = CIMClass(CLASSNAME)
        test_class.properties[PROPNAME] = test_prop
        self.conn.GetClass = Mock(return_value=test_class)

    def assertOutsideValueMap(self, vm, value):
        try:
            vm.tovalues(value)
        except ValueError as exc:
            if re.match("Element value outside of the set defined by "
                        "ValueMap.*", str(exc)) is None:
                self.fail("ValueError has unexpected text: %s" % str(exc))
        else:
            self.fail("ValueError was not raised.")

    def test_empty(self):
        valuemap = []
        values = []
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertOutsideValueMap(vm, 0)

    def test_zero(self):
        valuemap = ['0']
        values = ['zero']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(0), 'zero')
        self.assertOutsideValueMap(vm, 1)

    def test_one(self):
        valuemap = ['1']
        values = ['one']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertOutsideValueMap(vm, 0)
        self.assertEqual(vm.tovalues(1), 'one')
        self.assertOutsideValueMap(vm, 2)

    def test_one_cimtype(self):
        valuemap = ['1']
        values = ['one']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(Uint8(1)), 'one')

    def test_singles(self):
        valuemap = ['0', '1', '9']
        values = ['zero', 'one', 'nine']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(0), 'zero')
        self.assertEqual(vm.tovalues(1), 'one')
        self.assertOutsideValueMap(vm, 2)
        self.assertOutsideValueMap(vm, 8)
        self.assertEqual(vm.tovalues(9), 'nine')
        self.assertOutsideValueMap(vm, 10)

    def test_singles_ranges(self):
        valuemap = ['0', '1', '2..4', '..6', '7..', '9']
        values = ['zero', 'one', 'two-four', 'five-six', 'seven-eight', 'nine']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(0), 'zero')
        self.assertEqual(vm.tovalues(1), 'one')
        self.assertEqual(vm.tovalues(2), 'two-four')
        self.assertEqual(vm.tovalues(3), 'two-four')
        self.assertEqual(vm.tovalues(4), 'two-four')
        self.assertEqual(vm.tovalues(5), 'five-six')
        self.assertEqual(vm.tovalues(6), 'five-six')
        self.assertEqual(vm.tovalues(7), 'seven-eight')
        self.assertEqual(vm.tovalues(8), 'seven-eight')
        self.assertEqual(vm.tovalues(9), 'nine')
        self.assertOutsideValueMap(vm, 10)

    def test_unclaimed(self):
        valuemap = ['..']
        values = ['unclaimed']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(0), 'unclaimed')
        self.assertEqual(vm.tovalues(1), 'unclaimed')
        self.assertEqual(vm.tovalues(2), 'unclaimed')

    def test_singles_ranges_unclaimed(self):
        valuemap = ['0', '1', '2..4', '..6', '7..', '9', '..']
        values = ['zero', 'one', 'two-four', 'five-six', 'seven-eight', 'nine',
                  'unclaimed']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(0), 'zero')
        self.assertEqual(vm.tovalues(1), 'one')
        self.assertEqual(vm.tovalues(2), 'two-four')
        self.assertEqual(vm.tovalues(3), 'two-four')
        self.assertEqual(vm.tovalues(4), 'two-four')
        self.assertEqual(vm.tovalues(5), 'five-six')
        self.assertEqual(vm.tovalues(6), 'five-six')
        self.assertEqual(vm.tovalues(7), 'seven-eight')
        self.assertEqual(vm.tovalues(8), 'seven-eight')
        self.assertEqual(vm.tovalues(9), 'nine')
        self.assertEqual(vm.tovalues(10), 'unclaimed')
        self.assertEqual(vm.tovalues(11), 'unclaimed')

    def test_singles_ranges_unclaimed2(self):
        valuemap = ['0', '2..4', '..6', '7..', '9', '..']
        values = ['zero', 'two-four', 'five-six', 'seven-eight', 'nine',
                  'unclaimed']
        self.setup_for_property(valuemap, values)

        vm = ValueMapping.for_property(self.server, NAMESPACE, CLASSNAME,
                                       PROPNAME)

        self.assertEqual(vm.tovalues(0), 'zero')
        self.assertEqual(vm.tovalues(1), 'unclaimed')  # '..' fills this gap
        self.assertEqual(vm.tovalues(2), 'two-four')
        self.assertEqual(vm.tovalues(3), 'two-four')
        self.assertEqual(vm.tovalues(4), 'two-four')
        self.assertEqual(vm.tovalues(5), 'five-six')
        self.assertEqual(vm.tovalues(6), 'five-six')
        self.assertEqual(vm.tovalues(7), 'seven-eight')
        self.assertEqual(vm.tovalues(8), 'seven-eight')
        self.assertEqual(vm.tovalues(9), 'nine')
        self.assertEqual(vm.tovalues(10), 'unclaimed')
        self.assertEqual(vm.tovalues(11), 'unclaimed')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# (C) Copyright 2004,2005 Hewlett-Packard Development Company, L.P.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Tim Potter <tpot@hp.com>
#

"""
Validate XML instance data against the CIM-XML DTD.

Can also be invoked as a script for testing purposes, and then validates
XML data specified in standard input.
"""

from __future__ import print_function, absolute_import

import sys
import os
import os.path
from subprocess import Popen, PIPE, STDOUT

from pywbem.cim_obj import _ensure_bytes

DTD_FILE = 'CIM_DTD_V22.dtd'


def validate_xml(data, dtd_directory=None, root_elem=None):
    """
    Validate the provided XML instance data against a CIM-XML DTD, optionally
    requiring a particular XML root element.

    Arguments:

      * `data`: XML instance data to be validated.
      * `dtd_directory`: Directory with the DTD file (see `DTD_FILE` for name).
      * `root_elem`: Name of XML element that is expected as the root element
        in the XML instance data to be validated. None means no checking for
        a particular root element is performed.
    """

    dtd_file = DTD_FILE
    if dtd_directory is not None:
        dtd_file = os.path.join(dtd_directory, DTD_FILE).replace('\\', '/')

    # Make sure the XML data requires the specified root element, if any
    if root_elem is not None:
        data = '<!DOCTYPE %s SYSTEM "%s">\n' % (root_elem, dtd_file) + data
        xmllint_cmd = 'xmllint --valid --noout -'
    else:
        xmllint_cmd = 'xmllint --dtdvalid %s --noout -' % dtd_file

    p = Popen(xmllint_cmd, stdout=PIPE, stderr=STDOUT, stdin=PIPE, shell=True)

    data = _ensure_bytes(data)
    p.stdin.write(data)
    p.stdin.close()

    first_time = True
    for x in p.stdout.readlines():
        if first_time:
            first_time = False
            print("\nOutput from xmllint:")
        sys.stdout.write(x)

    status = p.wait()

    if status != 0:
        return False

    return True


if __name__ == '__main__':
    sys.exit(validate_xml(''.join(sys.stdin.readlines())))



#
# (C) Copyright 2017 InovaDevelopment Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

"""
Pywbem supports measuring the elapsed times of the WBEM operations that
were performed in context of a connection, and maintaining a statistics
over these times.

This capability is disabled by default and can be enabled in either
of these ways:

* When creating a :class:`~pywbem.WBEMConnection` object, via its
  ``enable_stats`` argument.

* After the :class:`~pywbem.WBEMConnection` object has been created, by
  modifying its :attr:`~pywbem.WBEMConnection.stats_enabled` instance
  attribute.

The :class:`~pywbem.Statistics` class maintains statistics over the measured
elapsed times of the WBEM operations and is the interface for accessing the
statistics. The statistics of a :class:`~pywbem.WBEMConnection` object are
accessible via its :attr:`~pywbem.WBEMConnection.statistics` instance
attribute.

The :class:`~pywbem.OperationStatistic` class is a helper class that contains
the actual measurement data for one operation name.
There will be one :class:`~pywbem.OperationStatistic` object each operation
name (see the table of WBEMConnection methods in the :ref:`WBEM operations`
section for the operation names).
The :class:`~pywbem.OperationStatistic` objects are under control of the
:class:`~pywbem.Statistics` class.

**Experimental:** The statistics support is experimental for this release.
"""

from __future__ import absolute_import

import time
import copy

__all__ = ['Statistics', 'OperationStatistic']


class OperationStatistic(object):
    # pylint: disable=too-many-instance-attributes
    """
    A statistics data keeper for the executions of all operations with the
    same operation name.

    This class maintains max, min, avg and count for elapsed time, request
    size, and response size over the executed operations of a single
    operation name.

    Use the :meth:`pywbem.Statistics.start_timer` method to create objects
    of this class::

        stats = container.start_timer('EnumerateInstances')
        ...
        stats.stop_timer(request_len, reply_len, server_time, exc)

    **Experimental:** This class is experimental for this release.
    """

    def __init__(self, container, name):
        """
        Parameters:

          container (:class:`~pywbem.Statistics`):
            The statistics container that holds this operation statistic
            object.

          name (string):
            Name of the operation.
        """
        self._container = container
        self._stat_start_time = None
        self._name = name
        self._count = 0
        self._exception_count = 0
        self._time_sum = float(0)
        self._time_min = float('inf')
        self._time_max = float(0)

        self._server_time_sum = float(0)
        self._server_time_min = float('inf')
        self._server_time_max = float(0)
        self._server_time_stored = False

        self._start_time = None

        self._request_len_sum = float(0)
        self._request_len_min = float('inf')
        self._request_len_max = float(0)

        self._reply_len_sum = float(0)
        self._reply_len_min = float('inf')
        self._reply_len_max = float(0)

    @property
    def stat_start_time(self):
        """
        :class:`py:float`: Point in time when the first statistic was taken
        since this object was either created or reset, in seconds since the
        epoch (for details, see :func:`py:time.time`).
        """
        return self._stat_start_time

    @property
    def name(self):
        """
        :term:`string`: Name of the operation for which this statistics
        object maintains data.

        This name is used by the :class:`~pywbem.Statistics` object
        holding this time statistics as a key.
        """
        return self._name

    @property
    def container(self):
        """
        :class:`~pywbem.Statistics`: The statistics container that holds
        this statistics object.
        """
        return self._container

    @property
    def count(self):
        """
        :term:`integer`: The number of measured operations (that is,
        invocations of the :meth:`~pywbem.OperationStatistic.stop_timer`
        method).
        """
        return self._count

    @property
    def exception_count(self):
        """
        :term:`integer`: The number of exceptions that occurred when invoking
        the measured operations.
        """
        return self._exception_count

    @property
    def avg_time(self):
        """
        :class:`py:float`: The average elapsed time for invoking the measured
        operations, in seconds.
        """
        try:
            return self._time_sum / self._count
        except ZeroDivisionError:
            return 0

    @property
    def min_time(self):
        """
        :class:`py:float`: The minimum elapsed time for invoking the measured
        operations, in seconds.
        """
        return self._time_min

    @property
    def max_time(self):
        """
        :class:`py:float`: The maximum elapsed time for invoking the measured
        operations, in seconds.
        """
        return self._time_max

    @property
    def avg_server_time(self):
        """
        :class:`py:float`: The average elapsed time for invoking the measured
        operations, in seconds.
        """
        try:
            return self._server_time_sum / self._count
        except ZeroDivisionError:
            return 0

    @property
    def min_server_time(self):
        """
        :class:`py:float`: The minimum elapsed time for invoking the measured
        operations, in seconds.
        """
        return self._server_time_min

    @property
    def max_server_time(self):
        """
        :class:`py:float`: The maximum elapsed time for invoking the measured
        operations, in seconds.
        """
        return self._server_time_max

    @property
    def avg_request_len(self):
        """
        :class:`py:float`: The average size of the HTTP body in the CIM-XML
        requests of the invoked operations, in Bytes.
        """
        try:
            return self._request_len_sum / self._count
        except ZeroDivisionError:
            return 0.0

    @property
    def min_request_len(self):
        """
        :class:`py:float`: The minimum size of the HTTP body in the CIM-XML
        requests of the invoked operations, in Bytes.
        """
        return self._request_len_min

    @property
    def max_request_len(self):
        """
        :class:`py:float`: The maximum size of the HTTP body in the CIM-XML
        requests of the invoked operations, in Bytes.
        """
        return self._request_len_max

    @property
    def avg_reply_len(self):
        """
        :class:`py:float`: The average size of the HTTP body in the CIM-XML
        responses of the invoked operations, in Bytes.
        """
        try:
            return self._reply_len_sum / self._count
        except ZeroDivisionError:
            return 0.0

    @property
    def min_reply_len(self):
        """
        :class:`py:float`: The minimum size of the HTTP body in the CIM-XML
        responses of the invoked operations, in Bytes.
        """
        return self._reply_len_min

    @property
    def max_reply_len(self):
        """
        :class:`py:float`: The maximum size of the HTTP body in the CIM-XML
        responses of the invoked operations, in Bytes.
        """
        return self._reply_len_max

    def reset(self):
        """
        Reset the statistics data for this object.
        """
        self._count = 0
        self._exception_count = 0
        self._stat_start_time = None
        self._time_sum = float(0)
        self._time_min = float('inf')
        self._time_max = float(0)

        self._server_time_sum = float(0)
        self._server_time_min = float('inf')
        self._server_time_max = float(0)
        self._server_time_stored = False

        self._request_len_sum = float(0)
        self._request_len_min = float('inf')
        self._request_len_max = float(0)

        self._reply_len_sum = float(0)
        self._reply_len_min = float('inf')
        self._reply_len_max = float(0)

    def start_timer(self):
        """
        This method needs to be called at the begin of an operation that
        is intended to be measured. It starts the measurement for that
        operation (by capturing the current point in time), if the
        statistics container holding this object is enabled. Otherwise,
        this method does nothing.

        A subsequent invocation of :meth:`~pywbem.OperationStatistic.stop_timer`
        will complete the measurement for that operation and will update the
        statistics data.
        """
        if self.container.enabled:
            self._start_time = time.time()
            if not self._stat_start_time:
                self._stat_start_time = self._start_time

    def stop_timer(self, request_len, reply_len, server_time=None,
                   exception=False):
        """
        This method needs to be called at the end of an operation that is
        intended to be measured. It completes the measurement for that
        operation by capturing the needed data, and updates the statistics
        data, if the statistics container holding this object is enabled.
        Otherwise, this method does nothing.

        Parameters:

          request_len (:term:`integer`)
            Size of the HTTP body of the CIM-XML request message, in Bytes.

          reply_len (:term:`integer`)
            Size of the HTTP body of the CIM-XML response message, in Bytes.

          exception (:class:`py:bool`)
            Boolean that specifies whether an exception was raised while
            processing the operation.

          server_time (:class:`py:bool`)
            Time in seconds that the server optionally returns to the
            client in the HTTP response defining the time from when the
            server received the request to when it started sending the
            response. If None, there is no time from the server.

        Returns:

          float: The elapsed time for the operation that just ended, or
          `None` if the statistics container holding this object is not
          enabled.
        """
        if self.container.enabled:
            if self._start_time is None:
                raise RuntimeError('stop_timer() called without preceding '
                                   ' start_timer()')
            dt = time.time() - self._start_time
            self._start_time = None
            self._count += 1
            self._time_sum += dt
            self._request_len_sum += request_len
            self._reply_len_sum += reply_len

            if exception:
                self._exception_count += 1

            if dt > self._time_max:
                self._time_max = dt
            if dt < self._time_min:
                self._time_min = dt

            if server_time:
                self._server_time_stored = True
                self._server_time_sum += server_time
                if dt > self._server_time_max:
                    self._server_time_max = server_time
                if dt < self._server_time_min:
                    self._server_time_min = server_time

            if request_len > self._request_len_max:
                self._request_len_max = request_len
            if request_len < self._request_len_min:
                self._request_len_min = request_len

            if reply_len > self._reply_len_max:
                self._reply_len_max = reply_len
            if reply_len < self._reply_len_min:
                self._reply_len_min = reply_len

            return dt
        else:
            return None

    def __repr__(self):
        """
        Return a human readable string with the statistics values, for debug
        purposes.
        """
        return 'OperationStatistic(' \
               'name={s.name!r}, ' \
               'count={s.count!r}, ' \
               'exception_count={s.exception_count!r}, ' \
               'avg_time={s.avg_time!r}, ' \
               'min_time={s.min_time!r}, ' \
               'max_time={s.max_time!r}, ' \
               'avg_server_time={s.avg_server_time!r}, ' \
               'min_server_time={s.min_server_time!r}, ' \
               'max_server_time={s.max_server_time!r}, ' \
               'avg_request_len={s.avg_request_len!r}, ' \
               'min_request_len={s.min_request_len!r}, ' \
               'max_request_len={s.max_request_len!r}, ' \
               'avg_reply_len={s.avg_reply_len!r}, ' \
               'min_reply_len={s.min_reply_len!r}, ' \
               'max_reply_len={s.max_reply_len!r})'. \
               format(s=self)

    #: Formatted header string (two lines), for use with the formatted rows
    #: returned by the :meth:`~pywbem.OperationStatistic.formatted` method.
    #:
    #: For an example, see :meth:`pywbem.Statistics.formatted`.
    formatted_header_w_svr = \
        'Count Excep            Time               ServerTime       ' \
        '       RequestLen               ReplyLen         Operation\n' \
        '        Cnt     Avg     Min     Max     Avg     Min     Max   ' \
        ' Avg    Min    Max      Avg      Min      Max\n'

    formatted_header = \
        'Count Excep            Time          ' \
        '       RequestLen            ReplyLen        Operation\n' \
        '        Cnt     Avg     Min     Max   ' \
        ' Avg    Min    Max    Avg      Min      Max\n'

    def formatted(self, include_server_time):
        """
        Return a formatted one-line string with the statistics values for this
        operation.

        The returned string can be used with the formatted header string
        defined in :attr:`~pywbem.OperationStatistic.formatted_header`.

        For an example, see :meth:`pywbem.Statistics.formatted`.
        """
        if include_server_time:
            return ('{0:5d} {1:5d} '
                    '{2:7.3f} {3:7.3f} {4:7.3f} '
                    '{5:7.3f} {6:7.3f} {7:7.3f} '
                    '{8:6.0f} {9:6.0f} {10:6.0f} '
                    '{11:8.0f} {12:8.0f} {13:8.0f} {14}\n'.
                    format(self.count, self.exception_count,
                           self.avg_time, self.min_time, self.max_time,
                           self.avg_server_time, self.min_server_time,
                           self.max_server_time,
                           self.avg_request_len, self.min_request_len,
                           self.max_request_len,
                           self.avg_reply_len, self.min_reply_len,
                           self.max_reply_len,
                           self.name))
        else:
            return ('{0:5d} {1:5d} '
                    '{2:7.3f} {3:7.3f} {4:7.3f} '
                    '{5:6.0f} {6:6.0f} {7:6.0f} '
                    '{8:6.0f} {9:8.0f} {10:8.0f} {11}\n'.
                    format(self.count, self.exception_count,
                           self.avg_time, self.min_time, self.max_time,
                           self.avg_request_len, self.min_request_len,
                           self.max_request_len,
                           self.avg_reply_len, self.min_reply_len,
                           self.max_reply_len,
                           self.name))


class Statistics(object):
    """
    A container class for multiple operation statistic objects (of class
    :class:`~pywbem.OperationStatistic`).

    Each operation statistic object is identified by a name, that is defined
    in the :meth:`~pywbem.Statistics.start_timer` method.

    The statistics container can be in a state of enabled or disabled.
    If enabled, it accumulates the elapsed times between subsequent calls to
    the :meth:`~pywbem.OperationStatistic.start_timer` and
    :meth:`~pywbem.OperationStatistic.stop_timer` methods of class
    :class:`~pywbem.OperationStatistic`.
    If disabled, calls to these methods do not accumulate any time.
    Initially, the statistics container is disabled. Its enablement state can
    be controlled via the :meth:`~pywbem.Statistics.enable` and
    :meth:`~pywbem.Statistics.disable` methods. Its current enablement
    state can be accessed via the :attr:`~pywbem.Statistics.enabled`
    property.

    **Experimental:** This class is experimental for this release.
    """

    def __init__(self, enable=False):
        """
        Parameters:

          enable (:class:`py:bool`):
            Initial enablement status for this statistics container.
        """
        # We convert any non-boolean values to True/False:
        self._enabled = bool(enable)
        self._op_stats = {}
        self._disabled_stats = OperationStatistic(self, "disabled")

    @property
    def enabled(self):
        """
        Indicates whether the statistics container is enabled.
        """
        return self._enabled

    def enable(self):
        """
        Enable the statistics container.
        """
        self._enabled = True

    def disable(self):
        """
        Disable the statistics container.
        """
        self._enabled = False

    def start_timer(self, name):
        """
        Start the timer for a given operation name and return the corresponding
        :class:`~pywbem.OperationStatistic` object, creating one if needed.

        Parameters:

          name (string):
            Name of the operation.

        Returns:

          :class:`~pywbem.OperationStatistic`: The operation statistic for the
          specified name. If this statistics container is disabled, a dummy
          operation statistic object is returned.
        """
        op_stat = self.get_op_statistic(name)
        op_stat.start_timer()
        return op_stat

    def get_op_statistic(self, name):
        """
        Get the :class:`~pywbem.OperationStatistic` object for an operation
        name or create a new object if an object for that name does not exist.

        Parameters:

          name (string):
            Name of the operation.

        Returns:

          :class:`~pywbem.OperationStatistic`: The operation statistic for the
          specified operation name. If this statistics container is disabled,
          a dummy operation statistic object is returned.
        """
        if not self.enabled:
            return self._disabled_stats
        if name not in self._op_stats:
            self._op_stats[name] = OperationStatistic(self, name)
        return self._op_stats[name]

    def snapshot(self):
        """
        Return a snapshot of the time statistics of this container.

        The snapshot represents the statistics data at the time this method
        is called, and remains unchanged even if the statistics of this
        container continues to be updated.

        Returns:

          : A list of tuples (name, stats) with:

          - name (:term:`string`): Name of the operation
          - stats (:class:`~pywbem.OperationStatistic`): Time statistics for
            the operation
        """
        return copy.deepcopy(self._op_stats).items()

    def __repr__(self):
        """
        Return a human readable display of the contents, for debug purposes.
        """
        ret = "Statistics(\n"
        for name in self._op_stats:
            ret += "  %r\n" % self._op_stats[name]
        ret += ")"
        return ret

    def formatted(self):
        # pylint: disable=line-too-long
        """
        Return a human readable string with the statistics for this container.
        The operations are sorted by decreasing average time.

        Server time statistic is included only if the wbem server has returned
        statistics information.

        Example w/o server times::

            Statistics (times in seconds, lengths in Bytes):
            Count  Exc            Time                ReqLen                ReplyLen          Operation
                   Cnt    Avg     Min     Max    Avg    Min    Max      Avg      Min      Max
                3    0   0.234   0.100   0.401   1233   1000   1500    26667    20000    35000 EnumerateInstances
                1    0   0.100   0.100   0.100   1200   1200   1200    22000    22000    22000 EnumerateInstanceNames
        """  # noqa: E501
        # pylint: enable=line-too-long
        ret = "Statistics (times in seconds, lengths in Bytes):\n"
        if self.enabled:
            snapshot = sorted(self.snapshot(),
                              key=lambda item: item[1].avg_time,
                              reverse=True)

            # Test to see if any server time is non-zero
            include_svr = False
            for name, stats in snapshot:  # pylint: disable=unused-variable
                if stats._server_time_stored:
                    include_svr = True
            if include_svr:
                ret += OperationStatistic.formatted_header_w_svr
            else:
                ret += OperationStatistic.formatted_header

            for name, stats in snapshot:  # pylint: disable=unused-variable
                ret += stats.formatted(include_svr)
        else:
            ret += "Disabled"
        return ret.strip()



#!/usr/bin/env python
"""
Test the complete PyWBEM client.

Approach:

* The HTTPretty package is used to intercept the client communication at the
  socket level. It catches PyWBEM requests going out and produces corresponding
  responses coming back in.

* Because HTTPretty seems to be able to register only one response for a
  particular URL, each test case can perform only one WBEM request and in its
  setup the test case needs to register the desired response for that request
  with HTTPretty.

* The testcases then performs the request at the PyWBEM client API, and
  verifies the response also at the level of the PyWBEM client API.

* HTTPretty remembers the last request that was sent, so after issuing the
  request, the CIM-XML produced by the PyWBEM client is verified based on that.
"""

from __future__ import print_function, absolute_import

import os
import sys
import doctest
import socket
import unittest
import re
import traceback
from collections import namedtuple
from io import open as _open
import six
import yaml
import httpretty
from httpretty.core import HTTPrettyRequestEmpty
from lxml import etree, doctestcompare
import pywbem
from pywbem.cim_obj import _ensure_unicode, NocaseDict
if six.PY2:
    from M2Crypto.Err import SSLError
else:
    from ssl import SSLError

# Directory with the JSON test case files, relative to this script:
TESTCASE_DIR = os.path.join(os.path.dirname(__file__), "testclient")

RUN_ONE_TESTCASE = None


def show_diff(conn, expected, actual, display_text):
    """Display the actual and expected data"""
    print("Details for the following assertion error:")
    print("- Expected result %s: %s" % (display_text, expected))
    print("- Actual result %s: %s" % (display_text, actual))
    if conn is not None and conn.debug:
        print("- HTTP response data: %r" % conn.last_raw_reply)


def str_tuple(tuple_):
    """
    Prepare a tuple or NonType for output.

    This gets around issues of type failure when trying to
    print tuples.

    Returns str rep of tuple on None if tuple_ is None
    """
    if tuple_ is None:
        return 'NoneType'
    else:
        return '%s' % (tuple_,)


class ClientTestError(Exception):
    """Exception indicating an issue in the test case definition."""
    pass


def obj(value, tc_name):
    """Create a PyWBEM CIM object or list of CIM Objects from a testcase value.
       If the input value is not a valid type or list of valid types an
       error is raised.
       If the pywbem_object defines a valid type (ex. CIMInstance) or
       list of items where the pywbem_object entry defines a valid type,
       the type is evaluated with the remaining entries in the
       dictionary and returned.

    Examples for a testcase value (JSON object / Python dict):

    1. CIMInstance object:

        {
            "pywbem_object": "CIMInstance",
            "classname": "PyWBEM_Person",
            "properties": {
                "Name": "Fritz",
                "Address": "Fritz Town"
            },
            "path": {
                "pywbem_object": "CIMInstanceName",
                "classname": "PyWBEM_Person",
                "namespace": "root/cimv2",
                "keybindings": {
                    "Name": "Fritz"
                }
            }
        }
        Returns a CIMInstance with the classname, properties, and path defined

    2. boolean object:

        False

        Returns boolean false

    3. list of objects
       A list of objects as defined by 1 above or boolean objects defined
       by 2. A list of corresponding results is returned

       Returns a list of the objects defined by each entry in the list

       See enumerateinstances.yaml for an example of list of instances.
       [] defines a null list

    """
    if isinstance(value, dict):
        if "pywbem_object" in value:
            ctor_name = value["pywbem_object"]
            try:
                ctor_call = getattr(pywbem, ctor_name)
            except AttributeError:
                raise ClientTestError("Error in definition of testcase %s: "
                                      "Unknown type specified in "
                                      "'pywbem_object' attribute: %s" %
                                      (tc_name, ctor_name))
            ctor_args = {}
            for arg_name in value:
                if arg_name == "pywbem_object":
                    continue
                ctor_args[arg_name] = obj(value[arg_name], tc_name)
            obj_ = ctor_call(**ctor_args)
        else:
            obj_ = {}
            for key in value:
                obj_[key] = obj(value[key], tc_name)
    elif isinstance(value, list):
        obj_ = [obj(x, tc_name) for x in value]
    else:
        obj_ = value
    return obj_


def tc_getattr(tc_name, dict_, key, default=-1):
    """ Gets the attribute of a name/value pair defined by key in dictionary
        defined by dict_.
        Get the attribute for the key and if the attribute is list or tuple
        only return the first item. If the attribute is a dictionary
        return the attribute
        If the key is not in the dictionary, return either the provided
        default or -1 if no default provided
    """
    try:
        value = dict_[key]
        if isinstance(value, (list, tuple)):
            value = value[0]
    except (KeyError, IndexError, TypeError):
        if default != -1:
            return default
        raise ClientTestError("Error in definition of testcase %s: "
                              "'%s' attribute missing" % (tc_name, key))
    return value


def tc_getattr_list(tc_name, dict_, key, default=-1):
    """ Gets the attribute of a name/value pair defined by key in dictionary
        defined by dict_. The entry may be dictionary, ordinary value or
        a list
        Gets the attribute for the key return it.  Does not test for
        type of the attribute.
        If the key is not in the dictionary, return either the provided
        default or -1 if no default provided

    """
    try:
        value = dict_[key]
    except KeyError:
        if default != -1:
            return default
        raise ClientTestError("Error in definition of testcase %s: "
                              "'%s' attribute missing" % (tc_name, key))
    return value


def tc_hasattr(dict_, key):
    """Return true if key is in dict_"""
    return key in dict_


class Callback(object):
    """A class with static methods that are HTTPretty callback functions for
    raising expected exceptions at the socket level.

    HTTPretty callback functions follow this interface:

        def my_callback(request, uri, headers):
            ...
            return (status, headers, body)

    Parameters:
      * `request`: string with invoked HTTP method
      * `uri`: string with target URI
      * `headers`: list of strings with HTTP headers of request

    Return value:
      * `status`: numeric with HTTP status code for response
      * `headers`: list of strings with HTTP headers for response
      * `body`: response body / payload

    They can also raise an exception, which is passed to the caller of the
    socket send call.
    """

    @staticmethod
    def socket_ssl(request, uri, headers):  # pylint: disable=unused-argument
        """HTTPretty callback function that raises an arbitrary
        SSLError."""
        raise SSLError(1, "Arbitrary SSL error.")

    @staticmethod
    def socket_104(request, uri, headers):  # pylint: disable=unused-argument
        """HTTPretty callback function that raises socket.error 104."""
        raise socket.error(104, "Connection reset by peer.")

    @staticmethod
    def socket_32(request, uri, headers):  # pylint: disable=unused-argument
        """HTTPretty callback function that raises socket.error 32."""
        raise socket.error(32, "Broken pipe.")

    @staticmethod
    def socket_timeout(request, uri, headers):  # pylint: disable=unused-argument # noqa: E501
        """HTTPretty callback function that raises socket.timeout error.
           The socket.timeout is just a string, no status.
        """
        raise socket.timeout("Socket timeout.")


def xml_embed(tree_elem):
    """
    Embed the CIM-XML representation of a CIM object.

    This is done by converting the XML tree representing the CIM object into a
    single XML element whose character text is the embedded representation,
    which has some special XML characters (defined in DSP0201) replaced with
    their character entity references (= XML escape sequences).

    Parameters:
      tree_elem (etree.Element):
        XML tree of the CIM-XML representation of the CIM object that is to
        be embedded.

    Returns:
      string: The embedded CIM-XML representation of the CIM object.
    """
    return xml_escape(etree.tostring(tree_elem))


def xml_escape(s):  # pylint: disable=invalid-name
    """
    Return the XML-escaped input string.
    """
    if isinstance(str, six.text_type):
        s = s.replace(u"&", u"&amp;")
        s = s.replace(u"<", u"&lt;")
        s = s.replace(u">", u"&gt;")
        s = s.replace(u"\"", u"&quot;")
        s = s.replace(u"'", u"&apos;")
    else:
        s = s.replace(b"&", b"&amp;")
        s = s.replace(b"<", b"&lt;")
        s = s.replace(b">", b"&gt;")
        s = s.replace(b"\"", b"&quot;")
        s = s.replace(b"'", b"&apos;")
    return s


def xml_unembed(emb_str):
    """
    Unembed the CIM-XML representation of a CIM object.

    This is done by converting an XML element whose character text is the
    embedded representation, into an XML tree representing the CIM object,
    whereby the character entity references (= XML escape sequences) of
    some special XML characters (defined in DSP0201) are replaced with
    their characters.

    Parameters:
      emb_str (string):
        The embedded CIM-XML representation of the CIM object that is to be
        unembedded.

    Returns:
      etree.Element:
        XML tree of the CIM-XML representation of the CIM object.
    """
    parser = etree.XMLParser(remove_blank_text=True)
    return etree.XML(xml_unescape(emb_str), parser=parser)


def xml_unescape(s):  # pylint: disable=invalid-name
    """
    Return the XML-unescaped input string.
    """
    if isinstance(s, six.text_type):
        s = s.replace(u"&lt;", u"<")
        s = s.replace(u"&gt;", u">")
        s = s.replace(u"&quot;", u"\"")
        s = s.replace(u"&apos;", u"'")
        s = s.replace(u"&amp;", u"&")
    else:
        s = s.replace(b"&lt;", b"<")
        s = s.replace(b"&gt;", b">")
        s = s.replace(b"&quot;", b"\"")
        s = s.replace(b"&apos;", b"'")
        s = s.replace(b"&amp;", b"&")
    return s


class EmbedUnembedTest(unittest.TestCase):
    """Test case for embed() / unembed() functions."""

    def test_unembed_simple(self):
        """Unembed a simple embedded instance string."""

        emb_str = b'&lt;INSTANCE NAME=&quot;C1&quot;&gt;' \
                  b'&lt;PROPERTY NAME=&quot;P1&quot;&gt;' \
                  b'&lt;VALUE&gt;V1&lt;/VALUE&gt;' \
                  b'&lt;/PROPERTY&gt;' \
                  b'&lt;/INSTANCE&gt;'

        instance_elem = xml_unembed(emb_str)

        self.assertEqual(instance_elem.tag, 'INSTANCE')
        self.assertEqual(len(instance_elem.attrib), 1)
        self.assertTrue('NAME' in instance_elem.attrib)
        self.assertEqual(instance_elem.attrib['NAME'], 'C1')

        self.assertEqual(len(instance_elem), 1)
        property_elem = instance_elem[0]
        self.assertEqual(property_elem.tag, 'PROPERTY')
        self.assertEqual(len(property_elem.attrib), 1)
        self.assertTrue('NAME' in property_elem.attrib)
        self.assertEqual(property_elem.attrib['NAME'], 'P1')

        self.assertEqual(len(property_elem), 1)
        value_elem = property_elem[0]
        self.assertEqual(value_elem.tag, 'VALUE')
        self.assertEqual(len(value_elem.attrib), 0)
        self.assertEqual(value_elem.text, 'V1')

    def test_embed_simple(self):
        """Embed a simple instance."""

        instance_elem = etree.Element('INSTANCE')
        instance_elem.attrib['NAME'] = 'C1'
        property_elem = etree.SubElement(instance_elem, 'PROPERTY')
        property_elem.attrib['NAME'] = 'P1'
        value_elem = etree.SubElement(property_elem, 'VALUE')
        value_elem.text = 'V1'

        emb_str = xml_embed(instance_elem)

        exp_emb_str = b'&lt;INSTANCE NAME=&quot;C1&quot;&gt;' \
                      b'&lt;PROPERTY NAME=&quot;P1&quot;&gt;' \
                      b'&lt;VALUE&gt;V1&lt;/VALUE&gt;' \
                      b'&lt;/PROPERTY&gt;' \
                      b'&lt;/INSTANCE&gt;'

        self.assertEqual(emb_str, exp_emb_str)


class ClientTest(unittest.TestCase):
    """Test case for PyWBEM client testing."""

    @staticmethod
    def assertXMLEqual(s_act, s_exp, entity=None):
        """Assert that the two XML fragments are equal, tolerating the following
        variations:
          * whitespace outside of element content and attribute values.
          * order of attributes.
          * order of certain child elements (see `sort_elements` in this
            function).

        Parameters:
          * s_act and s_exp are string representations of an XML fragment. The
            strings may be Unicode strings or UTF-8 encoded byte strings.
            The strings may contain an encoding declaration even when
            they are Unicode strings.

            Note: An encoding declaration is the `encoding` attribute in the
            XML declaration (aka XML processing statement), e.g.:
                <?xml version="1.0" encoding="utf-8" ?>
        """

        # Ensure Unicode strings and remove encoding from XML declaration
        encoding_pattern = re.compile(
            r'^<\?xml +(([a-zA-Z0-9_]+=".*")?) +' +
            r'encoding="utf-8" +(([a-zA-Z0-9_]+=".*")?) *\?>')
        encoding_repl = r'<?xml \1 \3 ?>'
        s_act = re.sub(encoding_pattern, encoding_repl, _ensure_unicode(s_act))
        s_exp = re.sub(encoding_pattern, encoding_repl, _ensure_unicode(s_exp))

        parser = etree.XMLParser(remove_blank_text=True)
        x_act = etree.XML(s_act, parser=parser)
        x_exp = etree.XML(s_exp, parser=parser)

        def sort_embedded(root, sort_elements):
            """
            Helper function for `sort_children()`, in support of embedded
            objects. This function invokes sort_children() on each embedded
            object in `root`, after unembedding the embedded object.

            Parameters:
              root (etree.Element):
                XML tree of the CIM-XML representation of the CIM element that
                contains an embedded CIM object (e.g. the CIM element may be
                an INSTANCE XML element and one of its PROPERTY child elements
                has a value that is an embedded CIM instance).
            """
            emb_elems = root.xpath("//*[@EmbeddedObject or @EMBEDDEDOBJECT]"
                                   "/*[local-name() = 'VALUE' or "
                                   "local-name() = 'VALUE.ARRAY']")
            for emb_elem in emb_elems:
                elem = xml_unembed(emb_elem.text)
                sort_children(elem, sort_elements)
                emb_elem.text = xml_embed(elem)

        def sort_children(root, sort_elements):
            """Sort certain elements in the `root` parameter to facilitate
               comparison of two XML documents.

               In addition, make sure this is also applied to any embedded
               objects (in their unembedded state).
            """
            sort_embedded(root, sort_elements)
            for tag, attr in sort_elements:
                # elems is a list of elements with this tag name
                elems = root.xpath("//*[local-name() = $tag]", tag=tag)
                if len(elems) > 0:
                    parent = elems[0].getparent()
                    first = None
                    after = None
                    for i in range(0, len(parent)):
                        if parent[i].tag == tag and first is None:
                            first = i
                        if parent[i].tag != tag and first is not None:
                            after = i
                    # The following changes the original XML tree:
                    # The following pylint warning can safely be disabled, see
                    # http://stackoverflow.com/a/25314665
                    # pylint: disable=cell-var-from-loop
                    parent[first:after] = sorted(elems,
                                                 key=lambda e: e.attrib[attr])

        sort_elements = [
            # Sort sibling elements with <first> tag by its <second> attribute
            ("IPARAMVALUE", "NAME"),
            ("PROPERTY", "NAME"),
            ("PROPERTY.ARRAY", "NAME"),
            ("PARAMETER", "NAME"),
            ("KEYBINDING", "NAME"),
        ]
        sort_children(x_act, sort_elements)
        sort_children(x_exp, sort_elements)

        ns_act = _ensure_unicode(etree.tostring(x_act))
        ns_exp = _ensure_unicode(etree.tostring(x_exp))

        checker = doctestcompare.LXMLOutputChecker()

        # This tolerates differences in whitespace and attribute order
        if not checker.check_output(ns_act, ns_exp, 0):
            diff = checker.output_difference(doctest.Example("", ns_exp),
                                             ns_act, 0)
            raise AssertionError("XML is not as expected in %s: %s" %
                                 (entity, diff))

    def test_all(self):
        """This method performs all tests of this script. It iterates through
        the JSON files in the test case directory and processes each of them.
        """

        print("")  # We are in the middle of test runner output

        # We need to work with absolute file paths, because this test may be
        # run from a different directory.
        for basefn in os.listdir(TESTCASE_DIR):
            relfn = os.path.join(TESTCASE_DIR, basefn)
            if relfn.endswith(".yaml"):
                self.runyamlfile(relfn)

    def runyamlfile(self, fn):
        """Read a YAML test case file and process each test case it defines.
        """

        print("Process YAML file: %s" % os.path.basename(fn))

        with _open(fn, encoding="utf-8") as fp:
            testcases = yaml.load(fp)
            for testcase in testcases:
                self.runtestcase(testcase)

    @httpretty.activate
    def runtestcase(self, testcase):
        """Run a single test case."""

        tc_name = tc_getattr("", testcase, "name")
        tc_desc = tc_getattr(tc_name, testcase, "description", None)
        tc_ignore = tc_getattr(tc_name, testcase, "ignore_python_version", None)
        tc_ignore_test = tc_getattr(tc_name, testcase, "ignore_test", None)

        # Test to determine if execute this testcase
        # 1. If the RUN_ONE_TESTCASE option set and this is not the one
        # 2. if ingore_python_version set and version does not match
        if RUN_ONE_TESTCASE is not None:
            if tc_name != RUN_ONE_TESTCASE:
                return
        else:
            if tc_ignore_test is not None:
                print("IGNORE  test case: %s: %s. Ignore_test: set" %
                      (tc_name, tc_desc))
                return
            if six.PY2 and tc_ignore == 2 or six.PY3 and tc_ignore == 3:
                print("IGNORE  test case: %s: %s for python version %s" %
                      (tc_name, tc_desc, tc_ignore))
                return

        print("Process test case: %s: %s" % (tc_name, tc_desc))

        httpretty.httpretty.allow_net_connect = False

        pywbem_request = tc_getattr(tc_name, testcase, "pywbem_request")
        exp_http_request = tc_getattr(tc_name, testcase, "http_request", None)
        http_response = tc_getattr(tc_name, testcase, "http_response", None)
        exp_pywbem_response = tc_getattr(tc_name, testcase, "pywbem_response")

        # Setup HTTPretty for one WBEM operation
        if exp_http_request is not None:
            exp_http_exception = tc_getattr(tc_name, http_response,
                                            "exception", None)
            if exp_http_exception is None:
                params = {
                    "body": tc_getattr(tc_name, http_response, "data"),
                    "adding_headers": tc_getattr(tc_name, http_response,
                                                 "headers", None),
                    "status": tc_getattr(tc_name, http_response, "status")
                }
            else:
                callback_name = exp_http_exception
                try:
                    callback_func = getattr(Callback(), callback_name)
                except AttributeError:
                    raise ClientTestError("Error in testcase %s: Unknown "
                                          "exception callback specified: %s" %
                                          (tc_name, callback_name))
                params = {
                    "body": callback_func
                }

            method = tc_getattr(tc_name, exp_http_request, "verb")
            uri = tc_getattr(tc_name, exp_http_request, "url")

            httpretty.register_uri(method=method, uri=uri, **params)

        conn = pywbem.WBEMConnection(
            url=tc_getattr(tc_name, pywbem_request, "url"),
            creds=tc_getattr(tc_name, pywbem_request, "creds"),
            default_namespace=tc_getattr(tc_name, pywbem_request, "namespace"),
            timeout=tc_getattr(tc_name, pywbem_request, "timeout"),
            enable_stats=tc_getattr(tc_name, pywbem_request, "enable-stats",
                                    False))

        conn.debug = tc_getattr(tc_name, pywbem_request, "debug", False)

        op = tc_getattr(tc_name, pywbem_request, "operation")
        # Example:
        #  "operation": {
        #      "pywbem_method": "GetInstance",
        #      "InstanceName": {
        #          "pywbem_object": "CIMInstanceName",
        #          "classname": "PyWBEM_Person",
        #          "keybindings": {
        #              "Name": "Fritz"
        #          }
        #      },
        #      "LocalOnly": False
        #  }

        op_name = tc_getattr(tc_name, op, "pywbem_method")
        op_args = {}
        for arg_name in op:
            if arg_name == "pywbem_method":
                continue
            op_args[arg_name] = obj(op[arg_name], tc_name)
        try:
            op_call = getattr(conn, op_name)

        except AttributeError as exc:
            raise ClientTestError("Error in definition of testcase %s: "
                                  "Unknown operation name: %s" %
                                  (tc_name, op_name))

        # Invoke the PyWBEM operation to be tested
        try:
            result = op_call(**op_args)
            raised_exception = None

        except Exception as exc:      # pylint: disable=broad-except
            raised_exception = exc
            stringio = six.StringIO()
            traceback.print_exc(file=stringio)
            raised_traceback_str = stringio.getvalue()
            stringio.close()
            result = None

        # Validate PyWBEM result and exceptions.
        # We validate exceptions before validating the HTTP request, because
        # an exception might have been raised on the way down before the
        # request was actually made.

        exp_exception = tc_getattr(tc_name, exp_pywbem_response,
                                   "exception", None)
        exp_cim_status = tc_getattr(tc_name, exp_pywbem_response,
                                    "cim_status", 0)

        # get the optional expected request and reply sizes if specified. The
        # default is None if not specified
        exp_request_len = tc_getattr_list(tc_name, exp_pywbem_response,
                                          "request_len", None)
        exp_reply_len = tc_getattr_list(tc_name, exp_pywbem_response,
                                        "reply_len", None)
        # get the expected result.  This may be either the the definition
        # of a value or cimobject or a list of values or cimobjects or
        # a named tuple of results.
        exp_result = tc_getattr_list(tc_name, exp_pywbem_response,
                                     "result", None)

        exp_pull_result = tc_getattr(tc_name, exp_pywbem_response,
                                     "pullresult", None)

        if exp_pull_result and exp_result:
            raise ClientTestError("Error in definition of testcase %s: "
                                  "result and pull result attributes "
                                  "are exclusive.")

        if exp_exception is not None and exp_result is not None:
            raise ClientTestError("Error in definition of testcase %s: "
                                  "'result' and 'exception' attributes in "
                                  "'pywbem_result' are not compatible." %
                                  tc_name)
        if exp_cim_status != 0 and exp_result is not None:
            raise ClientTestError("Error in definition of testcase %s: "
                                  "'result' and 'cim_status' attributes in "
                                  "'pywbem_result' are not compatible." %
                                  tc_name)

        if exp_cim_status != 0:
            exp_exception = 'CIMError'

        if exp_exception is not None:
            if raised_exception is None:
                raise AssertionError("Testcase %s: A %s exception was "
                                     "expected to be raised by PyWBEM "
                                     "operation %s, but no exception was "
                                     "actually raised." %
                                     (tc_name, exp_exception, op_name))
            elif raised_exception.__class__.__name__ != exp_exception:
                raise AssertionError("Testcase %s: A %s exception was "
                                     "expected to be raised by PyWBEM "
                                     "operation %s, but a different "
                                     "exception was actually raised:\n"
                                     "%s\n" %
                                     (tc_name, exp_exception, op_name,
                                      raised_traceback_str))
        else:
            if raised_exception is not None:
                raise AssertionError("Testcase %s: No exception was "
                                     "expected to be raised by PyWBEM "
                                     "operation %s, but an exception was "
                                     "actually raised:\n"
                                     "%s\n" %
                                     (tc_name, op_name, raised_traceback_str))

        # Validate HTTP request produced by PyWBEM

        if exp_http_request is not None:
            http_request = httpretty.last_request()
            self.assertTrue(not isinstance(http_request,
                                           HTTPrettyRequestEmpty),
                            "HTTP request is empty")
            exp_verb = tc_getattr(tc_name, exp_http_request, "verb")
            self.assertEqual(http_request.method, exp_verb,
                             "Verb in HTTP request is: %s (expected: %s)" %
                             (http_request.method, exp_verb))
            exp_headers = tc_getattr(tc_name, exp_http_request, "headers", {})
            for header_name in exp_headers:
                self.assertEqual(http_request.headers[header_name],
                                 exp_headers[header_name],
                                 "Value of %s header in HTTP request is: %s "
                                 "(expected: %s)" %
                                 (header_name,
                                  http_request.headers[header_name],
                                  exp_headers[header_name]))
            exp_data = tc_getattr(tc_name, exp_http_request, "data", None)
            self.assertXMLEqual(http_request.body, exp_data,
                                "Unexpected CIM-XML payload in HTTP request")
        if exp_request_len is not None:
            self.assertEqual(exp_request_len, conn.last_request_len,
                             'request lengths do not match. exp %s rcvd %s' %
                             (exp_request_len, conn.last_request_len))

            if conn.stats_enabled:
                snapshot = conn.statistics.snapshot()
                self.assertEqual(len(snapshot), 1)   # one operation; one stat

                for name, stats in snapshot:  # pylint: disable=unused-variable
                    stat = stats
                self.assertEqual(stat.count, 1, "Expected a single statistic")
                self.assertEqual(stat.min_request_len,
                                 stat.max_request_len)
                self.assertEqual(stat.min_request_len, exp_request_len)

        if exp_reply_len is not None:
            self.assertEqual(exp_reply_len, conn.last_reply_len,
                             'Reply lengths do not match. exp %s rcvd %s' %
                             (exp_reply_len, conn.last_reply_len))

            if conn.stats_enabled:
                snapshot = conn.statistics.snapshot()
                self.assertEqual(len(snapshot), 1)   # one operation; one stat

                for name, stats in snapshot:
                    stat = stats
                self.assertEqual(stat.count, 1, "Expected a single statistic")
                self.assertEqual(stat.min_reply_len,
                                 stat.max_reply_len)
                self.assertEqual(stat.min_reply_len, exp_reply_len)

        # Continue with validating the result

        if isinstance(raised_exception, pywbem.CIMError):
            cim_status = raised_exception.args[0]
        else:
            cim_status = 0
        self.assertEqual(cim_status, exp_cim_status,
                         "WBEMConnection operation CIM status code")

        # Returns either exp_result or exp_pull_result
        if exp_result is not None:
            exp_result_obj = obj(exp_result, tc_name)

            # The testcase can only specify lists but not tuples, so we
            # tolerate tuple/list mismatches:
            act_type = type(result)
            if act_type == tuple:
                act_type = list
            exp_type = type(exp_result_obj)
            # pylint: disable=unidiomatic-typecheck
            if act_type != exp_type:
                show_diff(None, type(exp_result_obj), type(result), 'type')
                raise AssertionError("PyWBEM CIM result type is not"
                                     " as expected.")

            # The testcase can only specify dicts but not NocaseDicts, so we
            # tolerate such mismatches (in case of InvokeMethod):
            if isinstance(exp_result_obj, list) and \
               len(exp_result_obj) == 2 and \
               isinstance(exp_result_obj[1], dict):
                _exp_result_obj = (
                    exp_result_obj[0],
                    NocaseDict(exp_result_obj[1])
                )
            else:
                # pylint: disable=redefined-variable-type
                _exp_result_obj = exp_result_obj

            # If result items are tuple, convert to lists. This is for
            # class ref and assoc results.
            if isinstance(result, list) and \
                    len(result) > 0 and isinstance(result[0], tuple):
                _result = []
                for item in result:
                    if isinstance(item, tuple):
                        _result.append(list(item))
                    else:
                        _result.append(item)
            else:
                # pylint: disable=redefined-variable-type
                _result = result

            if _result != _exp_result_obj:
                # TODO 2016/07 AM: Improve the presentation of the difference
                show_diff(conn, repr(exp_result_obj), repr(_result), 'data')
                raise AssertionError("WBEMConnection operation method result "
                                     "is not as expected.")

        # if this is a pull result, compare the components of expected
        # and actual results. Pull results return a tuple
        elif exp_pull_result is not None:
            exp_pull_result_obj = result_tuple(exp_pull_result, tc_name)

            # Result length should be the same as expected result
            if len(result) != len(exp_pull_result_obj):
                show_diff(conn, len(conn, exp_pull_result_obj), len(result),
                          'tuple size')
                raise AssertionError("PyWBEM CIM result type is not"
                                     " as expected.")
            # eos is required result
            if result.eos != exp_pull_result_obj.eos:
                show_diff(conn, exp_pull_result_obj.eos, result.eos,
                          'result.eos')
                raise AssertionError("WBEMConnection operation method result "
                                     "is not as expected.")

            # Context is required result
            # NOTE: pyaml does not natively support tuples. It supports very
            #   simple tuples but only with single objects and in block mode.
            exp_context = tuple(exp_pull_result_obj.context) \
                if exp_pull_result_obj.context \
                else None
            if result.context != exp_context:
                show_diff(conn, repr(str_tuple(exp_context)),
                          repr(str_tuple(result.context)), 'result.context')
                raise AssertionError("WBEMConnection operation method result "
                                     "is not as expected.")

            if "instances" in exp_pull_result:
                _result = result.instances
                _exp_result = exp_pull_result_obj.instances
            elif "paths" in exp_pull_result:
                _result = result.paths
                _exp_result = exp_pull_result_obj.paths
            else:
                raise AssertionError("WBEMConnection operation method result "
                                     "is not as expected. No 'instances' "
                                     "or 'paths' component.")

            if _result != _exp_result:
                # TODO 2016/07 AM: Improve the presentation of the diff.
                show_diff(conn, repr(_exp_result), repr(_result), 'result data')
                raise AssertionError("WBEMConnection operation method "
                                     "result is not as expected.")

            # TODO redo as indexed loop to compare all items.

        else:
            self.assertEqual(result, None,
                             "PyWBEM CIM result is not None: %s" % repr(result))


def result_tuple(value, tc_name):
    """ Process the value (a dictionary) to create a named tuple of
        the components that are part of a pull result.
        Returns the  namedtuple of either
            instance, eos, context
            or
            paths, eos, context
        For openqueryinstances it returns a named tuple with 4 elements
        in place of 3
    """
    if isinstance(value, dict):
        # test for both paths and instances.
        objs = None
        if "query_result_class" in value:
            result = namedtuple("exp_result", ["instances", "eos", "context",
                                               "query_result_class"])
        else:
            result = namedtuple("exp_result", ["instances", "eos", "context"])

        # either path or instances should be in value
        if "instances" in value:
            # instances = value["instances"]
            objs = obj(value["instances"], tc_name)
            if 'paths' in value:
                raise AssertionError("WBEMConnection operation method "
                                     "result is not as expected. Both "
                                     "'instances' and 'paths' component.")
        elif "paths" in value:
            # paths = value["paths"]
            objs = obj(value["paths"], tc_name)
            result = namedtuple("result", ["paths", "eos", "context"])
        else:
            raise AssertionError("WBEMConnection operation method result "
                                 "is not as expected. No 'instances' "
                                 "or 'paths' component.")

        # if query_result_class in value, add it to result
        if "query_result_class" in value:
            return result(objs, value["eos"], value["context"],
                          value["query_result_class"])
        else:
            return result(objs, value["eos"], value["context"])

    else:
        raise AssertionError("WBEMConnection operation invalid tuple "
                             "definition.")


if __name__ == '__main__':

    # NonDocumented option to run a single testcase if that testcase name
    # is listed on the cmd line.
    if len(sys.argv) > 1:
        print('Running a single testcase: %s' % sys.argv[1])
        RUN_ONE_TESTCASE = sys.argv[1]
        del sys.argv[1:2]

    elif len(sys.argv) > 2:
        sys.exit('ERROR: too many cmd line args')

    unittest.main()



#!/usr/bin/env python

# Copyright 2017 InovaDevelopment Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
"""
Unit test logging functionality in _logging.py

"""

from __future__ import absolute_import, print_function

import os
import logging

# Allows use of lots of single character variable names.
# pylint: disable=invalid-name,missing-docstring,too-many-statements
# pylint: disable=too-many-lines,no-self-use
import unittest
from testfixtures import LogCapture, log_capture

from pywbem import PywbemLoggers, LOG_OPS_CALLS_NAME
from pywbem._logging import get_logger

VERBOSE = False

# Location of any test scripts for testing wbemcli.py
SCRIPT_DIR = os.path.dirname(__file__)

LOG_FILE_NAME = 'test_logging.log'
TEST_OUTPUT_LOG = '%s/%s' % (SCRIPT_DIR, LOG_FILE_NAME)


class BaseLoggingTests(unittest.TestCase):
    """Base class for logging unit tests"""

    def setUp(self):
        # reset the PywbemLoggers store for the next test.  This forces
        # the dictionary back to empty. Since this is a class variable
        # we clear it at the class level.
        PywbemLoggers.loggers = {}

    def tearDown(self):
        LogCapture.uninstall_all()
        logging.shutdown()
        if os.path.isfile(TEST_OUTPUT_LOG):
            os.remove(TEST_OUTPUT_LOG)


class TestLogParse(BaseLoggingTests):
    """
    Test parse_log_specs
    """
    def parser_test(self, param, expected_result):
        """ Common test for successful parsing"""
        # pylint: disable=protected-access
        result = PywbemLoggers._parse_log_specs(param)
        self.assertEqual(result, expected_result)

    def test_comp_only(self):
        """Test all string"""
        param = 'all'
        self.parser_test(param, {'all': (None, None)})

    def test_comp_only2(self):
        """Test all= string"""
        param = 'all='
        self.parser_test(param, {'all': (None, None)})

    def test_complete1(self):
        """Test all=file:min string"""
        param = 'all=file:min'
        self.parser_test(param, {'all': ('file', 'min')})

    def test_complete2(self):
        """Test all=file:min: string"""
        param = 'all=file:min'
        self.parser_test(param, {'all': ('file', 'min')})

    def test_complete3(self):
        """Test all=file:min string"""
        param = 'all=file:min'
        self.parser_test(param, {'all': ('file', 'min')})

    def test_multiple1(self):
        """Test ops=file:min,http=file:min string"""
        param = 'ops=file:min,http=file:min'
        self.parser_test(param, {'ops': ('file', 'min'),
                                 'http': ('file', 'min')})

    def test_multiple2(self):
        """Test ops=file:min,http=file:min string"""
        param = 'ops=file:min,http='
        self.parser_test(param, {'ops': ('file', 'min'),
                                 'http': (None, None)})


class TestLogParseErrors(BaseLoggingTests):
    """ Test errors on the parse"""
    def parser_error_test(self, param):
        """Test for exception"""
        try:
            # pylint: disable=protected-access
            PywbemLoggers._parse_log_specs(param)
            self.fail('Param should generate exception %s' % param)
        except ValueError:
            pass

    def test_to_many_params(self):
        """ test all=file:min:junk string """
        param = "all=file:min:junk"
        self.parser_error_test(param)

    def test_empty(self):
        param = ""
        self.parser_error_test(param)


class TestLoggerCreate(BaseLoggingTests):
    """ Test the PywbemLoggers.create_logger method."""
    def test_create_single_logger1(self):
        """
        Create a simple logger
        """
        PywbemLoggers.create_logger('ops', 'file',
                                    log_filename=TEST_OUTPUT_LOG,
                                    log_detail_level='min')

        if VERBOSE:
            print('pywbem_loggers dict %s' % PywbemLoggers.loggers)
        expected_result = \
            {'pywbem.ops': ('min', 'file', TEST_OUTPUT_LOG)}

        # test getting from logger variable
        self.assertEqual(PywbemLoggers.loggers, expected_result)

        # test use of __repr__
        expected_result = 'PywbemLoggers(%s)' % expected_result
        self.assertEqual(('%r' % PywbemLoggers), expected_result)

    def test_create_single_logger2(self):
        """
        Create a simple logger from detailed parameter input
        """
        PywbemLoggers.create_logger('http', 'file',
                                    log_filename=TEST_OUTPUT_LOG,
                                    log_detail_level='min')

        if VERBOSE:
            print('pywbem_loggers dict %s' % PywbemLoggers.loggers)
        expected_result = \
            {'pywbem.http': ('min', 'file', TEST_OUTPUT_LOG)}

        self.assertEqual(PywbemLoggers.loggers, expected_result,
                         'Actual %s, Expected %s' % (PywbemLoggers.loggers,
                                                     expected_result))

    def test_create_single_logger3(self):
        """
        Create a simple logger from detailed parameter input
        """
        PywbemLoggers.create_logger('http', 'stderr',
                                    log_filename=None,
                                    log_detail_level='min')

        if VERBOSE:
            print('pywbem_loggers dict %s' % PywbemLoggers.loggers)
        expected_result = \
            {'pywbem.http': ('min', 'stderr', None)}

        self.assertEqual(PywbemLoggers.loggers, expected_result,
                         'Actual %s, Expected %s' % (PywbemLoggers.loggers,
                                                     expected_result))

    def test_create_single_logger4(self):
        """
        Create a simple logger from detailed parameter input
        """
        PywbemLoggers.create_logger('all', 'stderr',
                                    log_filename=None,
                                    log_detail_level='min')

        if VERBOSE:
            print('pywbem_loggers dict %s' % PywbemLoggers.loggers)
        expected_result = \
            {'pywbem.http': ('min', 'stderr', None),
             'pywbem.ops': ('min', 'stderr', None)}

        self.assertEqual(PywbemLoggers.loggers, expected_result)


class TestLoggerCreateErrors(BaseLoggingTests):
    """Test errors in the LoggerCreate Function"""

    def test_create_single_logger1(self):
        """
        Create a simple logger from detailed parameter input
        """
        try:
            PywbemLoggers.create_logger('httpx', 'stderr',
                                        log_filename=None,
                                        log_detail_level='min')
            self.fail('Exception expected')
        except ValueError as ve:
            if VERBOSE:
                print('ve %s' % ve)

    def test_create_single_logger2(self):
        """
        Create a simple logger from detailed parameter input
        """
        try:
            PywbemLoggers.create_logger('http', 'stderrblah',
                                        log_filename=None,
                                        log_detail_level='min')
            self.fail('Exception expected')
        except ValueError as ve:
            if VERBOSE:
                print('ve %s' % ve)

    def test_create_single_logger4(self):
        """
        Create a simple logger from detailed parameter input
        """
        try:
            PywbemLoggers.create_logger('http', 'stderr',
                                        log_filename=None,
                                        log_detail_level='mi')
            self.fail('Exception expected')
        except ValueError as ve:
            if VERBOSE:
                print('ve %s' % ve)

    def test_create_single_logger5(self):
        """
        Create a simple logger from detailed parameter input
        """
        try:
            PywbemLoggers.create_logger('http', 'file',
                                        log_filename=None,
                                        log_detail_level='mi')
            self.fail('Exception expected')
        except ValueError as ve:
            if VERBOSE:
                print('ve %s' % ve)


class TestLoggersCreate(BaseLoggingTests):
    """
    Tests to create loggers using the PywbemLogger class and the
    method create_loggers that creates logger definitions from
    an input string
    """
    def valid_loggers_create(self, input_str, expected_result,
                             log_filename=None):
        """Common test to do the create loggers and test result."""
        PywbemLoggers.create_loggers(input_str, log_filename)
        self.assertEqual(PywbemLoggers.loggers, expected_result)
        # TODO add test
        # for name in PywbemLoggers.loggers:
        #    print(PywbemLoggers.get_logger_info(name))

    def test_create_logger(self):
        """
        Create a simple logger
        """
        test_input = 'ops=file:min,http=file:min'

        expected_result = \
            {'pywbem.http': ('min', 'file', TEST_OUTPUT_LOG),
             'pywbem.ops': ('min', 'file', TEST_OUTPUT_LOG)}
        self.valid_loggers_create(test_input, expected_result,
                                  log_filename=TEST_OUTPUT_LOG)

    def test_create_loggers1(self):
        """
        Create a simple logger
        """
        test_input = 'all=file:min'
        expected_result = \
            {'pywbem.http': ('min', 'file', TEST_OUTPUT_LOG),
             'pywbem.ops': ('min', 'file', TEST_OUTPUT_LOG)}

        self.valid_loggers_create(test_input, expected_result,
                                  log_filename=TEST_OUTPUT_LOG)

    def test_create_loggers2(self):
        """
        Create a simple logger
        """
        test_input = 'all=stderr:all'
        expected_result = \
            {'pywbem.http': ('all', 'stderr', None),
             'pywbem.ops': ('all', 'stderr', None)}

        self.valid_loggers_create(test_input, expected_result)

    def test_get_logger_default(self):
        """ Test the get_logger function."""
        logger = get_logger(LOG_OPS_CALLS_NAME)
        self.assertEqual(len(logger.handlers), 1)
        log_info = PywbemLoggers.get_logger_info(LOG_OPS_CALLS_NAME)
        self.assertEqual(log_info[2], 'none')
        self.assertEqual(log_info[0], 'min')

    def test_get_logger_invalid(self):
        """ Test the get_logger function."""
        try:
            get_logger('pywbem.blah')
            self.fail('Expected exception for invalid log name')
        except ValueError:
            pass


class TestLoggerOutput(BaseLoggingTests):
    """Test output from logging"""

    @log_capture()
    def test_log_output(self, l):  # pylint: disable=blacklisted-name
        test_input = 'all=file:all'

        print('log filename %s' % TEST_OUTPUT_LOG)
        PywbemLoggers.create_loggers(test_input, TEST_OUTPUT_LOG)

        my_logger = get_logger(LOG_OPS_CALLS_NAME)

        self.assertNotEqual(my_logger, None,
                            'Valid named logger %s expected.'
                            % LOG_OPS_CALLS_NAME)
        # for name in PywbemLoggers.loggers:
        #    print(PywbemLoggers.get_logger_info(name))
        max_size = 1000
        result = 'This is fake return data'
        return_name = 'Return'
        if max_size and len(repr(result)) > max_size:
            result = '{:.{sz}s}...' \
                .format(repr(result), sz=max_size)
        else:
            result = '%s' % repr(result)

        my_logger.debug('%s: %s: %s', return_name, 'FakeMethodName', result)

        l.check(('pywbem.ops', 'DEBUG',
                 "Return: FakeMethodName: 'This is fake return data'"))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
"""
    Test to compare results of DOM tupletree implementation with the
    sax implementation
"""
import os
import xml
import unittest
import pprint
import six
from pkg_resources import resource_filename

from pywbem import tupletree

pp = pprint.PrettyPrinter(indent=4)

test_tuple = (u'CIM',
              {u'CIMVERSION': u'2.0', u'DTDVERSION': u'2.0'},
              [(u'MESSAGE',
                {u'ID': u'1001', u'PROTOCOLVERSION': u'1.0'},
                [u'\n',
                 (u'SIMPLERSP',
                  {},
                  [(u'IMETHODRESPONSE',
                    {u'NAME': u'Associators'},
                    [(u'IRETURNVALUE',
                      {},
                      [u'\n    \n',
                       (u'VALUE.OBJECTWITHPATH',
                        {},
                        [u'\n',
                         (u'INSTANCEPATH',
                          {},
                          [u'\n',
                           (u'NAMESPACEPATH',
                            {},
                            [u'\n',
                             (u'HOST',
                              {},
                              [u'10.10.10.10'],
                              None),
                             u'\n',
                             (u'LOCALNAMESPACEPATH',
                              {},
                              [u'\n',
                               (u'NAMESPACE',
                                {u'NAME': u'root'},
                                [],
                                None),
                               u'\n',
                               (u'NAMESPACE',
                                {u'NAME': u'emc'},
                                [],
                                None),
                               u'\n'],
                              None),
                             u'\n'],
                            None),
                           u'\n',
                           (u'INSTANCENAME',
                            {u'CLASSNAME': u'Symm_StorageVolume'},
                            [u'\n',
                             (u'KEYBINDING',
                              {u'NAME': u'CreationClassName'},
                              [(u'KEYVALUE',
                                {u'TYPE': u'string',
                                 u'VALUETYPE': u'string'},
                                [u'Symm_StorageVolume'],
                                None)],
                              None),
                             u'\n',
                             (u'KEYBINDING',
                              {u'NAME': u'DeviceID'},
                              [(u'KEYVALUE',
                                {u'TYPE': u'string',
                                 u'VALUETYPE': u'string'},
                                [u'00000'],
                                None)],
                              None),
                             u'\n',
                             (u'KEYBINDING',
                              {u'NAME': u'SystemCreationClassName'},
                              [(u'KEYVALUE',
                                {u'TYPE': u'string',
                                 u'VALUETYPE': u'string'},
                                [u'Symm_StorageSystem'],
                                None)],
                              None),
                             u'\n',
                             (u'KEYBINDING',
                              {u'NAME': u'SystemName'},
                              [(u'KEYVALUE',
                                {u'TYPE': u'string',
                                 u'VALUETYPE': u'string'},
                                [u'SYMMETRIX-+-000194900000'],
                                None)],
                              None),
                             u'\n'],
                            None),
                           u'\n'],
                          None),
                         u'\n',
                         (u'INSTANCE',
                          {u'CLASSNAME': u'Symm_StorageVolume'},
                          [u'\n',
                           (u'PROPERTY',
                            {u'NAME': u'Usage',
                             u'TYPE': u'uint16'},
                            [(u'VALUE',
                              {},
                              [u'3'],
                              None),
                             u'\n'],
                            None),
                           u'\n'],
                          None),
                         u'\n'],
                        None),
                       u'\n\n'],
                      None)],
                    None)],
                  None)],
                None)],
              None)


# The following was originally the pywbem parser.  Now it is uses simply
# as a check on the sax parser.  Therefore, this code was moved from
# tupletree.py to here for the tests. Nov 2016. Note that this code is
# referenced from other code in the attic so we might not want to eliminate
# in completely.
def dom_to_tupletree(node):
    """Convert a DOM object to a pyRXP-style tuple tree.

    Each element is a 4-tuple of (NAME, ATTRS, CONTENTS, None).

    Very nice for processing complex nested trees.
    """

    if node.nodeType == node.DOCUMENT_NODE:
        # boring; pop down one level
        return dom_to_tupletree(node.firstChild)
    assert node.nodeType == node.ELEMENT_NODE

    name = node.nodeName
    attrs = {}
    contents = []

    for child in node.childNodes:
        if child.nodeType == child.ELEMENT_NODE:
            contents.append(dom_to_tupletree(child))
        elif child.nodeType == child.TEXT_NODE:
            assert isinstance(child.nodeValue, six.string_types), \
                "text node %s is not a string %r" % child
            contents.append(child.nodeValue)
        elif child.nodeType == child.CDATA_SECTION_NODE:
            contents.append(child.nodeValue)
        else:
            raise RuntimeError("can't handle %r" % child)

    for i in range(node.attributes.length):
        attr_node = node.attributes.item(i)
        attrs[attr_node.nodeName] = attr_node.nodeValue

    # XXX: Cannot handle comments, cdata, processing instructions, etc.

    # it's so easy in retrospect!
    return (name, attrs, contents, None)


def xml_to_tupletree(xml_string):
    """
    Parse XML straight into tupletree.
    Uses the minidom to parse xml_string int a dom object.
    This is part of the old obsolete dom code
    """
    dom_xml = xml.dom.minidom.parseString(xml_string)
    return dom_to_tupletree(dom_xml)


class TestTupleTreeRegression(unittest.TestCase):
    """Setup the TupleTree tests"""
    def setUp(self):
        self.data_dir = resource_filename(__name__, 'tupletree_ok')
        for path, dirs, files in os.walk(self.data_dir):
            self.filenames = [os.path.join(path, f) for f in files]

    def test_to_tupletree(self):
        """
        SAX parsing replicates DOM parsing to the tupletree.

        This test compares the input xml compiled with the dom and sax.

        """
        for fn in self.filenames:
            with open(fn, 'rb') as fh:
                xml_str = fh.read()
            tree_dom = xml_to_tupletree(xml_str)
            tree_sax = tupletree.xml_to_tupletree_sax(xml_str)
            self.assertEqual(tree_dom, tree_sax,
                             'SAX and DOM not equal for %s:\n%s,\n%s'
                             % (fn, pp.pformat(tree_sax), pp.pformat(tree_dom)))


class TestTupleTreeError(unittest.TestCase):

    def setUp(self):
        self.data_dir = resource_filename(__name__, 'tupletree_error')

    def test_to_tupletree_error(self):
        """SAX parsing generates errors."""
        filename = os.path.join(self.data_dir, 'Associators_error.xml')
        with open(filename, 'rb') as fh:
            xml_str = fh.read()
            self.assertRaises(xml.sax.SAXParseException,
                              tupletree.xml_to_tupletree_sax, xml_str)


class TestTupleTreeSax(unittest.TestCase):

    def test_xml_to_tupletree_sax(self):
        """
        XML to tupletree with SAX is accurate.

        This test compares data to the test_tuple variable in this file.
        """
        data_dir = resource_filename(__name__, 'tupletree_ok')
        path = os.path.join(data_dir, 'Associators_StorageVolume_small.xml')
        with open(path, 'rb') as fh:
            xml_str = fh.read()
        tree_sax = tupletree.xml_to_tupletree_sax(xml_str)
        self.assertEqual(test_tuple, tree_sax,
                         'Test tuple and SAX not equal for %s:\n%s,\n%s'
                         % (path, pp.pformat(tree_sax), pp.pformat(tree_sax)))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
    Test wbemcli script.  This test generates a cmdline that calls
    wbemcli with a specific set of options and tests the returns.
    Because wbemcli always goes to interactive mode, the test call
    includes a wbemcli script that forces wbemcli to quit.

    It dynamically generates the set of tests from the TEST_MAP table
    where each test is a call to execute wbemcli with a particular set of
    arguments and options. It then tests the stdout, stderr, and exitcode
    against the TEST_MAP.
"""

from __future__ import print_function, absolute_import
import os
import unittest
import re
from subprocess import Popen, PIPE
from collections import namedtuple
import six

# Output fragments to test against for each test defined
# Each item is a list of fragmants that are tested against the cmd execution
# result
HELP_OUTPUT = ['-n namespace, --namespace',
               '-t timeout, --timeout']
LOG_DEST_STDERR_OUTPUT = ['log=on',
                          'Connection: http://blah,',
                          ' no creds']
LOG_DEST_FILE_OUTPUT = ['log=on',
                        'Connection: http://blah,',
                        ' no creds']

TIMEOUT_OUTPUT = ['Connection: http://blah,',
                  ' no creds',
                  'timeout=10']

STATS_OUTPUT = ['Connection: http://blah,',
                ' no creds',
                'stats=on']
NAMESPACE_OUTPUT = ['log=off',
                    'Connection: http://blah,',
                    ' no creds',
                    'verifycert=on',
                    'default-namespace=rex/fred']
LOG_STDERR_OUTPUT = ['pywbem.ops-Connection:']

DEF_NAMESPACE_OUTPUT = ['default-namespace=root/cimv2']


# pylint: disable=invalid-name
tst_def = namedtuple('tst_def', ['test_name',
                                 'cmd',
                                 'expected_stdout',
                                 'expected_exitcode',
                                 'expected_stderr',
                                 'url'])

# Each test in the following list is a namedtuple (tst_def) containing test
# definition.
TESTS_MAP = [  # pylint: disable=invalid-name
    tst_def('help', '--help', HELP_OUTPUT, 0, None, None),
    tst_def('namespace', '-n rex/fred', NAMESPACE_OUTPUT, 0, None, None),
    tst_def('timeout', '-t 10', TIMEOUT_OUTPUT, 0, None, None),
    tst_def('log_dest_file1', '-l ops=file', LOG_DEST_FILE_OUTPUT, 0, None,
            None),
    tst_def('log_dest_file1a', '-l ops=file:min', LOG_DEST_FILE_OUTPUT, 0,
            None, None),
    tst_def('log_dest_file2', '-l http=file', LOG_DEST_FILE_OUTPUT, 0, None,
            None),
    tst_def('log_dest_file3', '-l all=file', LOG_DEST_FILE_OUTPUT, 0, None,
            None),
    tst_def('log_dest_file4', '-l ops=stderr:all', LOG_DEST_FILE_OUTPUT,
            0, LOG_STDERR_OUTPUT, None),
    tst_def('def_namespace', '', DEF_NAMESPACE_OUTPUT, 0, None, None),
    tst_def('error_param', '-n', None, 2,
            ['argument -n/--namespace: expected one argument'], None),
    tst_def('error_url', '', None, 2,
            ['error: Invalid scheme on server argument'], 'httpx://xxx'), ]


def create_abs_path(filename):
    """
    create an absolute path name for filename in the same directory as this
    python code.  Also, replace any backslashes with forward slashes to
    account for executing this in windows.
    """
    script_dir = os.path.dirname(__file__)
    script_file = os.path.join(script_dir, filename)
    return script_file.replace('\\', '/')


class ContainerMeta(type):
    """Metaclass to define the function to generate unittest methods."""

    def __new__(mcs, name, bases, dict):  # pylint: disable=redefined-builtin
        def generate_test(test_name, test_params):
            """
            Defines the test method (test_<name>) that we generate for each test
            and returns the method.

            The cmd_str defines ONLY the arguments and options part of the
            command.  This function prepends wbemcli to the cmd_str.

            Since wbemcli is interactive, it also includes a quit script

            Each test builds the pywbemcli command executes it and tests the
            results
            """

            def test(self):  # pylint: disable=missing-docstring
                """ The test method that is generated."""
                # create the path for the quit script. Required to
                # exit wbemcli after the test.
                quit_script_file = create_abs_path('wbemcli_quit_script.py')

                script_name = 'wbemcli.bat' if os.name == 'nt' else 'wbemcli'

                url = 'http://blah' if test_params.url is None \
                    else test_params.url

                cmd = ('%s %s %s -s %s' % (script_name,
                                           url,
                                           test_params.cmd,
                                           quit_script_file))

                # Disable python warnings for wbemcli call
                # because some imports generate deprecated warnings
                # that appear in std_err when nothing expected
                bash_cmd = 'bash -c "set -o pipefail;  PYTHONPATH=. ' \
                           'PYTHONWARNINGS= %s"' % cmd

                proc = Popen(bash_cmd, shell=True, stdout=PIPE, stderr=PIPE)
                std_out, std_err = proc.communicate()
                exitcode = proc.returncode

                # print('bash_cmd: %s\nexitcode: %s std_out:\n%s\nstd_err:'
                #      '\n%s' % (bash_cmd, exitcode, std_out, std_err))

                if six.PY3:
                    std_out = std_out.decode()
                    std_err = std_err.decode()

                if test_params.expected_exitcode is not None:
                    self.assertEqual(exitcode, test_params.expected_exitcode,
                                     ('Test %s: Unexpected ExitCode Err. '
                                      'Expected %s: cmd="%s": '
                                      'exitcode %s: stderr=%s' %
                                      (test_name, test_params.expected_exitcode,
                                       cmd,
                                       exitcode, std_err)))

                if test_params.expected_stderr is None:
                    if re.search('ImportWarning', std_err) is None:
                        self.assertEqual(std_err, "",
                                         'Test %s stderr not empty as '
                                         'expected. Returned %s'
                                         % (test_name, std_err))
                    else:
                        print('Ignored junk in stderr %s' % std_err)
                    self.assertEqual(std_err, "",
                                     'Test %s stderr not empty as expected. '
                                     'Returned %s'
                                     % (test_name, std_err))
                else:
                    for item in test_params.expected_stderr:
                        match_result = re.search(item, std_err)
                        self.assertNotEqual(match_result, None, 'Test %s, '
                                            'stderr did not match test '
                                            'definition. Expected %s in %s' %
                                            (test_name, item, std_err))

                if test_params.expected_stdout is not None:
                    for item in test_params.expected_stdout:
                        match_result = re.search(item, std_out)
                        self.assertNotEqual(match_result, None,
                                            'Test=%s, stdout did not match '
                                            'test definition. Expected %s in %s'
                                            % (test_name, item, std_out))
                else:
                    self.assertEqual(std_out, "",
                                     'Test %s stdout not empty as expected. '
                                     'Returned %s'
                                     % (test_name, std_out))
            return test

        # generate individual unittest functions from TESTS_MAP list
        for test_params in TESTS_MAP:
            test_name = "test_%s" % test_params.test_name
            dict[test_name] = generate_test(test_name, test_params)
        return type.__new__(mcs, name, bases, dict)


@six.add_metaclass(ContainerMeta)
class TestsContainer(unittest.TestCase):
    """Container class for all tests created from ContainerMeta"""
    __metaclass__ = ContainerMeta


if __name__ == '__main__':
    unittest.main()



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
Operation recorder interface and implementations.
"""

from __future__ import print_function, absolute_import
from collections import namedtuple

try:
    from collections import OrderedDict  # pylint: disable=import-error
except ImportError:
    from ordereddict import OrderedDict  # pylint: disable=import-error

from datetime import datetime, timedelta
import logging
import yaml
from yaml.representer import RepresenterError
import six

from .cim_obj import CIMInstance, CIMInstanceName, CIMClass, CIMClassName, \
    CIMProperty, CIMMethod, CIMParameter, CIMQualifier, \
    CIMQualifierDeclaration, NocaseDict
from .cim_types import CIMInt, CIMFloat, CIMDateTime
from .exceptions import CIMError
from ._logging import PywbemLoggers, LOG_OPS_CALLS_NAME, LOG_HTTP_NAME
from .config import DEFAULT_MAX_LOG_ENTRY_SIZE

if six.PY2:
    import codecs  # pylint: disable=wrong-import-order


__all__ = ['BaseOperationRecorder', 'TestClientRecorder',
           'LogOperationRecorder',
           'OpArgs', 'OpResult', 'HttpRequest', 'HttpResponse']

if six.PY2:
    _Longint = long  # noqa: F821
else:
    _Longint = int

OpArgsTuple = namedtuple("OpArgsTuple", ["method", "args"])


def _represent_ordereddict(dump, tag, mapping, flow_style=None):
    """PyYAML representer function for OrderedDict.
    This is needed for yaml.safe_dump() to support OrderedDict.

    Courtesy:
    http://blog.elsdoerfer.name/2012/07/26/make-pyyaml-output-an-ordereddict/
    """
    value = []
    node = yaml.MappingNode(tag, value, flow_style=flow_style)
    if dump.alias_key is not None:
        dump.represented_objects[dump.alias_key] = node
    best_style = True
    if hasattr(mapping, 'items'):
        mapping = mapping.items()
    for item_key, item_value in mapping:
        node_key = dump.represent_data(item_key)
        node_value = dump.represent_data(item_value)
        if not (isinstance(node_key, yaml.ScalarNode) and
                not node_key.style):
                    best_style = False    # pylint: disable=bad-indentation
        if not (isinstance(node_value, yaml.ScalarNode) and
                not node_value.style):
                    best_style = False    # pylint: disable=bad-indentation
        value.append((node_key, node_value))
    if flow_style is None:
        if dump.default_flow_style is not None:
            node.flow_style = dump.default_flow_style
        else:
            node.flow_style = best_style
    return node


yaml.SafeDumper.add_representer(
    OrderedDict,
    lambda dumper, value:
    _represent_ordereddict(dumper, u'tag:yaml.org,2002:map', value))


# Some monkey-patching for better diagnostics:
def _represent_undefined(self, data):
    """Raises flag for objects that cannot be represented"""
    raise RepresenterError("cannot represent an object: %s of type: %s; "
                           "yaml_representers: %r, "
                           "yaml_multi_representers: %r" %
                           (data, type(data), self.yaml_representers.keys(),
                            self.yaml_multi_representers.keys()))


yaml.SafeDumper.represent_undefined = _represent_undefined


class OpArgs(OpArgsTuple):
    """
    A named tuple representing the name and input arguments of the invocation
    of a :class:`~pywbem.WBEMConnection` method, with the following named fields
    and attributes:

    Attributes:

      method (:term:`unicode string`):
        Name of the :class:`~pywbem.WBEMConnection` method.

      args (:class:`py:dict`):
        Dictionary of input arguments (both positional and keyword-based).
    """
    __slots__ = ()

    def __repr__(self):
        return "OpArgs(method={s.method!r}, args={s.args!r})".format(s=self)


OpResultTuple = namedtuple("OpResultTuple", ["ret", "exc"])


class OpResult(OpResultTuple):
    """
    A named tuple representing the result of the invocation of a
    :class:`~pywbem.WBEMConnection` method, with the following named fields
    and attributes:

    Attributes:

      ret (:class:`py:object`):
        Return value, if the method returned.
        `None`, if the method raised an exception.

        Note that `None` may be a legitimate return value, so the test for
        exceptions should be done based upon the :attr:`exc` variable.

      exc (:exc:`~py:exceptions.Exception`):
        Exception object, if the method raised an exception.
        `None`, if the method returned.
    """
    __slots__ = ()

    def __repr__(self):
        return "OpResult(ret={s.ret!r}, exc={s.exc!r})".format(s=self)


HttpRequestTuple = namedtuple("HttpRequestTuple",
                              ["version", "url", "target", "method", "headers",
                               "payload"])


class HttpRequest(HttpRequestTuple):
    """
    A named tuple representing the HTTP request sent by the WBEM client, with
    the following named fields and attributes:

    Attributes:

      version (:term:`number`):
        HTTP version from the request line (10 for HTTP/1.0, 11 for HTTP/1.1).

      url (:term:`unicode string`):
        URL of the WBEM server (e.g. 'https://myserver.acme.com:15989').

      target (:term:`unicode string`):
        Target URL segment as stated in request line (e.g. '/cimom').

      method (:term:`unicode string`):
        HTTP method as stated in the request line (e.g. "POST").

      headers (:class:`py:dict`):
        A dictionary of all HTTP header fields:

        * key (:term:`unicode string`): Name of the header field
        * value (:term:`unicode string`): Value of the header field

      payload (:term:`unicode string`):
        HTTP payload, i.e. the CIM-XML string.
    """
    __slots__ = ()

    def __repr__(self):
        return "HttpRequest(version={s.version!r}, url={s.url!r}, " \
               "target={s.target!r}, method={s.method!r}, " \
               "headers={s.headers!r}, payload={s.payload!r})" \
               .format(s=self)


HttpResponseTuple = namedtuple("HttpResponseTuple",
                               ["version", "status", "reason", "headers",
                                "payload"])


class HttpResponse(HttpResponseTuple):
    """
    A named tuple representing the HTTP response received by the WBEM client,
    with the following named fields and attributes:

    Attributes:

      version (:term:`number`):
        HTTP version from the response line (10 for HTTP/1.0, 11 for HTTP/1.1).

      status (:term:`number`):
        HTTP status code from the response line (e.g. 200).

      reason (:term:`unicode string`):
        HTTP reason phrase from the response line (e.g. "OK").

      headers (:class:`py:dict`):
        A dictionary of all HTTP header fields:

        * key (:term:`unicode string`): Name of the header field
        * value (:term:`unicode string`): Value of the header field

      payload (:term:`unicode string`):
        HTTP payload, i.e. the CIM-XML string.
    """
    __slots__ = ()

    def __repr__(self):
        return "HttpResponse(version={s.version!r}, status={s.status!r}, " \
               "reason={s.reason!r}, headers={s.headers!r}, " \
               "payload={s.payload!r})".format(s=self)


class BaseOperationRecorder(object):
    # pylint: disable=too-many-instance-attributes
    """
    Abstract base class defining the interface to an operation recorder,
    that records the WBEM operations executed in a connection to a WBEM
    server.

    An operation recorder can be registered by setting the
    :attr:`~pywbem.WBEMConnection.operation_recorder` instance
    attribute of the :class:`~pywbem.WBEMConnection` object to an
    object of a subclass of this base class.

    When an operation recorder is registered on a connection, each operation
    that is executed on the connection will cause the :meth:`record`
    method of the operation recorder object to be called, if the recorder is
    enabled.

    The operation recorder is by default enabled, and can be disabled and
    re-enabled using the :meth:`~pywbem.BaseOperationRecorder.disable` and
    :meth:`~pywbem.BaseOperationRecorder.enable` methods, respectively.
    This can be used to temporarily pause the recorder.
    """

    def __init__(self):
        self._enabled = True
        self._conn_id = None
        self.reset()

    def enable(self):
        """Enable the recorder."""
        self._enabled = True

    def disable(self):
        """Disable the recorder."""
        self._enabled = False

    @property
    def enabled(self):
        """Indicate whether the recorder is enabled."""
        return self._enabled

    @staticmethod
    def open_file(filename, file_mode='w'):
        """
        A static convience function that performs the open of the recorder file
        correctly for different versions of python.  This covers the
        issue where the file should be opened in text mode but that is
        done differently in python 2 and python 3

        Parameters:

          filename(:term:`string`):
            Name of the file where the recorder output will be written

          file_mode(:term:`string`):
            Optional file mode.  The default is 'w' which overwrites any
            existing file.  if 'a' is used, the data is appended to any
            existing file.

        Example::

            recorder = TestClientRecorder(
                BaseOperationRecorder.open_file('recorder.log'))
        """
        if six.PY2:
            # Open with codecs to define text mode
            return codecs.open(filename, mode=file_mode, encoding='utf-8')

        return open(filename, file_mode, encoding='utf8')

    def reset(self, pull_op=None):
        """Reset all the attributes in the class. This also allows setting
        the pull_op attribute that defines whether the operation is to be
        a traditional or pull operation.
        This does NOT reset _conn.id as that exists through the life of
        the connection.
        """
        self._pywbem_method = None
        self._pywbem_args = None

        self._pywbem_result_ret = None
        self._pywbem_result_exc = None

        self._http_request_version = None
        self._http_request_url = None
        self._http_request_target = None
        self._http_request_method = None
        self._http_request_headers = None
        self._http_request_payload = None

        self._http_response_version = None
        self._http_response_status = None
        self._http_response_reason = None
        self._http_response_headers = None
        self._http_response_payload = None
        self._pull_op = pull_op

    def stage_wbem_connection(self, wbem_connection):
        """
        Stage information about the connection. Used only by
        LogOperationRecorder.
        """
        pass

    def stage_pywbem_args(self, method, **kwargs):
        """
        Set requst method and all args.
        Normally called before the cmd is executed to record request
        parameters
        """
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_method = method
        self._pywbem_args = kwargs

    def stage_pywbem_result(self, ret, exc):
        """ Set Result return info or exception info"""
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_result_ret = ret
        self._pywbem_result_exc = exc

    def stage_http_request(self, conn_id, version, url, target, method, headers,
                           payload):
        """Set request HTTP information including url, headers, etc."""
        # pylint: disable=attribute-defined-outside-init
        self._http_request_version = version
        self._http_request_conn_id = conn_id
        self._http_request_url = url
        self._http_request_target = target
        self._http_request_method = method
        self._http_request_headers = headers
        self._http_request_payload = payload

    # pylint: disable=unused-argument
    def stage_http_response1(self, conn_id, version, status, reason, headers):
        """Set response http info including headers, status, etc.
           conn_id unused here. Used in log"""
        # pylint: disable=attribute-defined-outside-init
        self._http_response_version = version
        self._http_response_status = status
        self._http_response_reason = reason
        self._http_response_headers = headers

    def stage_http_response2(self, payload):
        """Stage second part of http response, the payload"""
        # pylint: disable=attribute-defined-outside-init
        self._http_response_payload = payload

    def record_staged(self):
        """Encode staged information on request and result to output"""
        if self.enabled:
            pwargs = OpArgs(
                self._pywbem_method,
                self._pywbem_args)
            pwresult = OpResult(
                self._pywbem_result_ret,
                self._pywbem_result_exc)
            httpreq = HttpRequest(
                self._http_request_version,
                self._http_request_url,
                self._http_request_target,
                self._http_request_method,
                self._http_request_headers,
                self._http_request_payload)
            httpresp = HttpResponse(
                self._http_response_version,
                self._http_response_status,
                self._http_response_reason,
                self._http_response_headers,
                self._http_response_payload)
            self.record(pwargs, pwresult, httpreq, httpresp)

    def record(self, pywbem_args, pywbem_result, http_request, http_response):
        """
        Function that is called to record a single WBEM operation, i.e. the
        invocation of a single :class:`~pywbem.WBEMConnection` method.

        This function is called only when the recorder is enabled, i.e. it
        does not need to check for recorder enablement.

        Parameters:

          pywbem_args (:class:`~pywbem.OpArgs`):
            The name and input arguments of the :class:`~pywbem.WBEMConnection`
            method that is recorded.

          pywbem_result (:class:`~pywbem.OpResult`):
            The result (return value or exception) of the
            :class:`~pywbem.WBEMConnection` method that is recorded.

          http_request (:class:`~pywbem.HttpRequest`):
            The HTTP request sent by the :class:`~pywbem.WBEMConnection` method
            that is recorded.

            `None`, if no HTTP request had been sent (e.g. because an exception
            was raised before getting there).

          http_response (:class:`~pywbem.HttpResponse`):
            The HTTP response received by the :class:`~pywbem.WBEMConnection`
            method that is recorded.

            `None`, if no HTTP response had been received (e.g. because an
            exception was raised before getting there).
        """
        raise NotImplementedError


class LogOperationRecorder(BaseOperationRecorder):
    """
    An Operation Recorder that logs the information to a set of named logs.
    This recorder uses two named logs:

    LOG_OPS_CALLS_NAME - Logger for cim_operations method calls and responses

    LOG_HTTP_NAME - Logger for http_requests and responses

    This also implements a method to log information on each connection.

    All logging calls are at the debug level.
    """
    def __init__(self, max_log_entry_size=None):
        """
        Creates the the loggers and sets the max_log_size for each if
        the input parameter max_log_entry_size is not `None`.

        Parameters: (:term:`integer`)

          max_log_entry_size(:term:`integer`)
            The maximum size of each log entry. This is primarily to limit
            response sizes since they could be enormous.
            If `None`, no size limit and the full request or response is logged.
        """
        super(LogOperationRecorder, self).__init__()

        # compute max entry size for each logger
        max_sz = max_log_entry_size if max_log_entry_size \
            else DEFAULT_MAX_LOG_ENTRY_SIZE

        self.opslogger = logging.getLogger(LOG_OPS_CALLS_NAME)
        ops_logger_info = PywbemLoggers.get_logger_info(LOG_OPS_CALLS_NAME)
        opsdetaillevel = ops_logger_info[0] if ops_logger_info else None

        self.ops_max_log_size = max_sz if opsdetaillevel == 'min'  \
            else None

        self.httplogger = logging.getLogger(LOG_HTTP_NAME)
        http_logger_info = PywbemLoggers.get_logger_info(LOG_HTTP_NAME)
        httpdetaillevel = http_logger_info[0] if http_logger_info else None
        self.http_max_log_size = max_sz if httpdetaillevel == 'min' \
            else None

    def stage_wbem_connection(self, wbem_connection):
        """
        Log connection information. This includes the connection id
        that should remain throught the life of the connection.
        """
        self._conn_id = wbem_connection.conn_id

        if self.enabled:
            self.opslogger.debug('Connection:%s %r', self._conn_id,
                                 wbem_connection)

    def stage_pywbem_args(self, method, **kwargs):
        """
        Log request method and all args.
        Normally called before the cmd is executed to record request
        parameters.
        This method does not limit size of log record.
        """
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_method = method
        if self.enabled and self.opslogger.isEnabledFor(logging.DEBUG):
            # Order kwargs.  Note that this is done automatically starting
            # with python 3.6
            kwstr = ', '.join([('{0}={1!r}'.format(key, kwargs[key]))
                               for key in sorted(six.iterkeys(kwargs))])

            self.opslogger.debug('Request:%s %s(%s)', self._conn_id, method,
                                 kwstr)

    def stage_pywbem_result(self, ret, exc):
        """
        Log result return or exception parameter. This function allows
        setting maximum size on the result parameter logged because response
        information can be very large
        .
        """
        def format_result(ret, max_len):
            """ format ret as repr while clipping it to max_len if
                max_len is not None.
            """
            result = '{0!r}'.format(ret)
            if max_len and (len(result) > max_len):
                result = (result[:max_len] + '...')
            return result

        if self.enabled and self.opslogger.isEnabledFor(logging.DEBUG):
            if exc:  # format exception
                result = format_result(
                    '%s(%s)' % (exc.__class__.__name__, exc),
                    self.ops_max_log_size)
            else:    # format result
                # test if type is tuple (subclass of tuple but not type tuple)
                # pylint: disable=unidiomatic-typecheck
                if isinstance(ret, tuple) and type(ret) is not tuple:
                    try:    # test if field instances or paths
                        rtn_data = ret.instances
                        data_str = 'instances'
                    except AttributeError:
                        rtn_data = ret.paths
                        data_str = 'paths'
                    rtn_data = format_result(rtn_data, self.ops_max_log_size)

                    try:    # test for query_result_class
                        qrc = ', query_result_class=%s' % ret.query_result_class
                    except AttributeError:
                        qrc = ""

                    result = "{0.__name__}(context={1}, eos={2}{3}, {4}={5})" \
                        .format(type(ret), ret.context, ret.eos, qrc,
                                data_str, rtn_data)
                else:
                    result = format_result(ret, self.ops_max_log_size)

            return_type = 'Exception' if exc else 'Return'

            self.opslogger.debug('%s:%s %s(%s)', return_type, self._conn_id,
                                 self._pywbem_method,
                                 result)

    def stage_http_request(self, conn_id, version, url, target, method, headers,
                           payload):
        """Log request HTTP information including url, headers, etc."""
        if self.enabled and self.httplogger.isEnabledFor(logging.DEBUG):
            # pylint: disable=attribute-defined-outside-init
            # if Auth header, mask data
            if 'Authorization' in headers:
                authtype, cred = headers['Authorization'].split(' ')
                headers['Authorization'] = '%s %s' % (authtype, 'X' * len(cred))

            header_str = ' '.join('{0}:{1!r}'.format(k, v)
                                  for k, v in headers.items())

            self.httplogger.debug('Request:%s %s %s %s %s %s\n    %s',
                                  conn_id, method, target, version, url,
                                  header_str, payload)

    def stage_http_response1(self, conn_id, version, status, reason, headers):
        """Set response http info including headers, status, etc. """
        # pylint: disable=attribute-defined-outside-init
        self._http_response_version = version
        self._http_response_status = status
        self._http_response_reason = reason
        self._http_response_headers = headers
        self._http_response_conn_id = conn_id

    def stage_http_response2(self, payload):
        """Log complete http response, including response1 and payload"""

        # required because http code uses sending all None to reset
        # parameters. We ignore that
        if not self._http_response_version and not payload:
            return
        if self.enabled and self.httplogger.isEnabledFor(logging.DEBUG):
            if self._http_response_headers:
                header_str = \
                    ' '.join('{0}:{1!r}'.format(k, v)
                             for k, v in self._http_response_headers.items())
            else:
                header_str = ''

            # format the payload possibly with max size limit
            payload = '%r' % payload.decode('utf-8')
            if self.http_max_log_size and \
                    (len(payload) > self.http_max_log_size):
                payload = (payload[:self.http_max_log_size] + '...')

            self.httplogger.debug('Response:%s %s:%s %s %s\n    %s',
                                  self._http_response_conn_id,
                                  self._http_response_status,
                                  self._http_response_reason,
                                  self._http_response_version,
                                  header_str,
                                  payload)

    def record_staged(self):
        """Not used for logging"""
        pass

    def record(self, pywbem_args, pywbem_result, http_request, http_response):
        """Not used for logging"""
        pass


class TestClientRecorder(BaseOperationRecorder):
    """
    An operation recorder that generates test cases for each recorded
    operation. The test cases are in the YAML format suitable for the
    `test_client` unit test module of the pywbem project.
    """

    # HTTP header fields to exclude when creating the testcase
    # (in lower case)
    EXCLUDE_REQUEST_HEADERS = [
        'authorization',
        'content-length',
        'content-type',
    ]
    EXCLUDE_RESPONSE_HEADERS = [
        'content-length',
        'content-type',
    ]

    # Dummy server URL and credentials for use in generated test case
    TESTCASE_URL = 'http://acme.com:80'
    TESTCASE_USER = 'username'
    TESTCASE_PASSWORD = 'password'

    def __init__(self, fp):
        """
        Parameters:

          fp (file):
            An open file that each test case will be written to.  This file
            should have been opened in text mode.

            Since there are differences between python 2 and 3 in opening
            files in text mode, the static method
            :meth:`~pywbem.BaseOperationRecorder.open_file`
            can be used to open the file or python 2/3 compatible open::

              from io import open
                  f = open('blah.log', encoding='utf-8')

        Example::

          recorder = TestClientRecorder(
              BaseOperationRecorder.open_file('recorder.log'))
        """
        super(TestClientRecorder, self).__init__()
        self._fp = fp

    def record(self, pywbem_args, pywbem_result, http_request, http_response):
        """
        Function that records the invocation of a single
        :class:`~pywbem.WBEMConnection` method, by appending a corresponding
        test case to the file.

        Parameters: See :meth:`pywbem.BaseOperationRecorder.record`.
        """

        testcase = OrderedDict()
        testcase['name'] = pywbem_args.method
        testcase['description'] = 'Generated by TestClientRecorder'

        tc_pywbem_request = OrderedDict()
        tc_pywbem_request['url'] = TestClientRecorder.TESTCASE_URL
        tc_pywbem_request['creds'] = [TestClientRecorder.TESTCASE_USER,
                                      TestClientRecorder.TESTCASE_PASSWORD]
        tc_pywbem_request['namespace'] = 'root/cimv2'
        tc_pywbem_request['timeout'] = 10
        tc_pywbem_request['debug'] = False
        tc_operation = OrderedDict()
        tc_operation['pywbem_method'] = pywbem_args.method
        for arg_name in pywbem_args.args:
            tc_operation[arg_name] = self.toyaml(pywbem_args.args[arg_name])
        tc_pywbem_request['operation'] = tc_operation
        testcase['pywbem_request'] = tc_pywbem_request

        tc_pywbem_response = OrderedDict()
        if pywbem_result.ret is not None:
            yaml_txt = 'pullresult' if self._pull_op else 'result'
            tc_pywbem_response[yaml_txt] = self.toyaml(pywbem_result.ret)

        if pywbem_result.exc is not None:
            exc = pywbem_result.exc
            if isinstance(exc, CIMError):
                tc_pywbem_response['cim_status'] = self.toyaml(exc.status_code)
            else:
                tc_pywbem_response['exception'] = self.toyaml(
                    exc.__class__.__name__)
        testcase['pywbem_response'] = tc_pywbem_response

        tc_http_request = OrderedDict()
        if http_request is not None:
            tc_http_request['verb'] = http_request.method
            tc_http_request['url'] = TestClientRecorder.TESTCASE_URL
            if http_request.target:
                tc_http_request['url'] += http_request.target
            tc_request_headers = OrderedDict()
            if http_request.headers is not None:
                for hdr_name in http_request.headers:
                    if hdr_name.lower() not in \
                       TestClientRecorder.EXCLUDE_REQUEST_HEADERS:
                        tc_request_headers[hdr_name] = \
                            http_request.headers[hdr_name]
            tc_http_request['headers'] = tc_request_headers
            if http_request.payload is not None:
                data = http_request.payload.decode('utf-8')
                data = data.replace('><', '>\n<').strip()
            else:
                data = None
            tc_http_request['data'] = data
        testcase['http_request'] = tc_http_request

        tc_http_response = OrderedDict()
        if http_response is not None:
            tc_http_response['status'] = http_response.status
            tc_response_headers = OrderedDict()
            if http_response.headers is not None:
                for hdr_name in http_response.headers:
                    if hdr_name.lower() not in \
                       TestClientRecorder.EXCLUDE_RESPONSE_HEADERS:
                        tc_response_headers[hdr_name] = \
                            http_response.headers[hdr_name]
            tc_http_response['headers'] = tc_response_headers
            if http_response.payload is not None:
                data = http_response.payload.decode('utf-8')
                data = data.replace('><', '>\n<').strip()
            else:
                data = None
            tc_http_response['data'] = data
        else:
            tc_http_response['exception'] = "# Change this to a callback " \
                                            "function that causes this " \
                                            "condition."
        testcase['http_response'] = tc_http_response

        testcases = []
        testcases.append(testcase)

        # The file is open in text mode, so we produce a unicode string
        data = yaml.safe_dump(testcases, encoding=None, allow_unicode=True,
                              default_flow_style=False, indent=4)
        data = data.replace('\n\n', '\n')  # YAML dump duplicates newlines
        self._fp.write(data)
        self._fp.flush()

    def toyaml(self, obj):
        """
        Convert any allowable input argument to or return value from an
        operation method to an object that is ready for serialization into
        test_client yaml format.
        """

        # namedtuple is subclass of tuple so it is instance of tuple but
        # not type tuple. Cvt to dictionary and cvt dict to yaml.
        # pylint: disable=unidiomatic-typecheck
        if isinstance(obj, tuple) and type(obj) is not tuple:
            ret_dict = obj._asdict()
            return self.toyaml(ret_dict)
        if isinstance(obj, (list, tuple)):
            ret = []
            # This does not handle namedtuple
            for item in obj:
                ret.append(self.toyaml(item))
            return ret
        elif isinstance(obj, (dict, NocaseDict)):
            ret_dict = OrderedDict()
            for key in obj.keys():  # get keys in original case
                ret_dict[key] = self.toyaml(obj[key])
            return ret_dict
        elif obj is None:
            return obj
        elif isinstance(obj, six.binary_type):
            return obj.decode("utf-8")
        elif isinstance(obj, six.text_type):
            return obj
        elif isinstance(obj, CIMInt):
            return _Longint(obj)
        elif isinstance(obj, (bool, int)):
            # TODO ks jun 17 should the above be six.integertypes???
            # The check for int must be after CIMInt, because CIMInt is int.
            return obj
        elif isinstance(obj, CIMFloat):
            return float(obj)
        elif isinstance(obj, CIMDateTime):
            return str(obj)
        elif isinstance(obj, datetime):
            return CIMDateTime(obj)
        elif isinstance(obj, timedelta):
            return CIMDateTime(obj)
        elif isinstance(obj, CIMInstance):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMInstance'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['properties'] = self.toyaml(obj.properties)
            ret_dict['path'] = self.toyaml(obj.path)
            return ret_dict
        elif isinstance(obj, CIMInstanceName):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMInstanceName'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['namespace'] = self.toyaml(obj.namespace)
            ret_dict['keybindings'] = self.toyaml(obj.keybindings)
            return ret_dict
        elif isinstance(obj, CIMClass):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMClass'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['superclass'] = self.toyaml(obj.superclass)
            ret_dict['properties'] = self.toyaml(obj.properties)
            ret_dict['methods'] = self.toyaml(obj.methods)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMClassName):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMClassName'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['host'] = self.toyaml(obj.host)
            ret_dict['namespace'] = self.toyaml(obj.namespace)
            return ret_dict
        elif isinstance(obj, CIMProperty):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMProperty'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['value'] = self.toyaml(obj.value)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['reference_class'] = self.toyaml(obj.reference_class)
            ret_dict['embedded_object'] = self.toyaml(obj.embedded_object)
            ret_dict['is_array'] = self.toyaml(obj.is_array)
            ret_dict['array_size'] = self.toyaml(obj.array_size)
            ret_dict['class_origin'] = self.toyaml(obj.class_origin)
            ret_dict['propagated'] = self.toyaml(obj.propagated)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMMethod):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMMethod'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['return_type'] = self.toyaml(obj.return_type)
            ret_dict['class_origin'] = self.toyaml(obj.class_origin)
            ret_dict['propagated'] = self.toyaml(obj.propagated)
            ret_dict['parameters'] = self.toyaml(obj.parameters)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMParameter):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMParameter'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['reference_class'] = self.toyaml(obj.reference_class)
            ret_dict['is_array'] = self.toyaml(obj.is_array)
            ret_dict['array_size'] = self.toyaml(obj.array_size)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMQualifier):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMQualifier'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['value'] = self.toyaml(obj.value)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['propagated'] = self.toyaml(obj.propagated)
            ret_dict['tosubclass'] = self.toyaml(obj.tosubclass)
            ret_dict['toinstance'] = self.toyaml(obj.toinstance)
            ret_dict['overridable'] = self.toyaml(obj.overridable)
            ret_dict['translatable'] = self.toyaml(obj.translatable)
            return ret_dict
        elif isinstance(obj, CIMQualifierDeclaration):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMQualifierDeclaration'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['value'] = self.toyaml(obj.value)
            ret_dict['is_array'] = self.toyaml(obj.is_array)
            ret_dict['array_size'] = self.toyaml(obj.array_size)
            ret_dict['scopes'] = self.toyaml(obj.scopes)
            ret_dict['tosubclass'] = self.toyaml(obj.tosubclass)
            ret_dict['toinstance'] = self.toyaml(obj.toinstance)
            ret_dict['overridable'] = self.toyaml(obj.overridable)
            ret_dict['translatable'] = self.toyaml(obj.translatable)
            return ret_dict
        else:
            raise TypeError("Invalid type in TestClientRecorder.toyaml(): "
                            "%s %s" % (obj.__class__.__name__, type(obj)))



"""
Extensions for unittest module.

The ideas for these extensions are based on functions from the comfychair
module by Martin Pool <mbp@samba.org>.
"""

from __future__ import absolute_import

import re
import os

from pywbem.cim_obj import NocaseDict


class RegexpMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for regular expression tests."""

    def assertRegexpContains(self, text, pattern, msg=None):
        """Assert that a string text *contains* a particular regular expression
        pattern.

        Parameters:
            text         string: string text to be searched
            pattern      string: regular expression

        Raises:
            AssertionError if not matched
          """
        if not re.search(pattern, text):
            raise AssertionError(
                "String text does not contain regexp pattern\n"
                "    text:    %r\n"
                "    pattern: %r" % (text, pattern))

    def assertRegexpMatches(self, text, pattern, msg=None):
        """Assert that a string text matches a particular regular expression
        pattern.

        Note: This method is provided by unittest starting from Python 2.7.
        For Python 2.6, we need this function here.

        Parameters:
            text         string: string text to be matched
            pattern      string: regular expression

        Raises:
            AssertionError if not matched
          """
        if not re.match(pattern, text):
            raise AssertionError(
                "String text does not match regexp pattern\n"
                "    text:    %r\n"
                "    pattern: %r" % (text, pattern))


class FileMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for file related tests."""

    def assertFileExists(self, filename):
        assert os.path.exists(filename), \
            ("File does not exist but should: %s" % filename)

    def assertFileNotExists(self, filename):
        assert not os.path.exists(filename), \
            ("File exists but should not: %s" % filename)


class CIMObjectMixin(object):
    """Mixin class for classes derived from unittest.TestCase, providing
    assertion functions for CIM object related tests."""

    def assertCIMInstanceName(self, obj, classname, keybindings, host=None,
                              namespace=None):
        """
        Verify the attributes of the CIMInstanceName object in `obj` against
        expected values passed as the remaining arguments.
        """
        self.assertEqual(obj.classname, classname, "classname attribute")
        self.assertEqual(obj.keybindings, keybindings, "keybindings attribute")
        self.assertEqual(obj.host, host, "host attribute")
        self.assertEqual(obj.namespace, namespace, "namespace attribute")

    def assertCIMProperty(self, obj, name, value, type_=None,
                          class_origin=None, array_size=None, propagated=None,
                          is_array=False, reference_class=None, qualifiers=None,
                          embedded_object=None):
        """
        Verify the attributes of the CIMProperty object in `obj` against
        expected values passed as the remaining arguments.
        """
        self.assertEqual(obj.name, name, "name attribute")
        self.assertEqual(obj.value, value, "value attribute")
        self.assertEqual(obj.type, type_, "type attribute")
        self.assertEqual(obj.class_origin, class_origin,
                         "class_origin attribute")
        self.assertEqual(obj.array_size, array_size, "array_size attribute")
        self.assertEqual(obj.propagated, propagated, "propagated attribute")
        self.assertEqual(obj.is_array, is_array, "is_array attribute")
        self.assertEqual(obj.reference_class, reference_class,
                         "reference_class attribute")
        self.assertEqual(obj.qualifiers, NocaseDict(qualifiers),
                         "qualifiers attribute")
        self.assertEqual(obj.embedded_object, embedded_object,
                         "embedded_object attribute")

    def assertEqualCIMClass(self, act_class, exp_class):
        """
        Verify that two CIMClass objects are equal.
        """
        self.assertEqual(act_class.classname,
                         exp_class.classname,
                         "Unexpected classname: %s" % act_class.classname)
        context = "class %s" % act_class.classname
        self.assertEqual(act_class.superclass,
                         exp_class.superclass,
                         "Unexpected superclass: %s" % act_class.superclass)
        self.assertEqualProperties(
            act_class.properties, exp_class.properties,
            context)
        self.assertEqualMethods(
            act_class.methods, exp_class.methods,
            context)
        self.assertEqualQualifiers(
            act_class.qualifiers, exp_class.qualifiers,
            context)

    def assertEqualProperties(self, act_properties, exp_properties, context):
        """
        Verify that two dicts of CIMProperty objects are equal.
        """

        self.assertEqualKeys(act_properties,
                             exp_properties,
                             "%s, property names" % context)
        for propname in act_properties:
            self.assertEqualCIMProperty(
                act_properties[propname], exp_properties[propname],
                "%s, property %s" % (context, propname))

    def assertEqualMethods(self, act_methods, exp_methods, context):
        """
        Verify that two dicts of CIMMethod objects are equal.
        """

        self.assertEqualKeys(act_methods,
                             exp_methods,
                             "%s, method names" % context)
        for methname in act_methods:
            self.assertEqualCIMMethod(
                act_methods[methname], exp_methods[methname],
                "%s, method %s" % (context, methname))

    def assertEqualParameters(self, act_parameters, exp_parameters, context):
        """
        Verify that two dicts of CIMParameter objects are equal.
        """

        self.assertEqualKeys(act_parameters,
                             exp_parameters,
                             "%s, parameter names" % context)
        for parmname in act_parameters:
            self.assertEqualCIMParameter(
                act_parameters[parmname], exp_parameters[parmname],
                "%s, parameter %s" % (context, parmname))

    def assertEqualQualifiers(self, act_qualifiers, exp_qualifiers, context):
        """
        Verify that two dicts of CIMQualifier objects are equal.
        """

        self.assertEqualKeys(act_qualifiers,
                             exp_qualifiers,
                             "%s, qualifier names" % context)
        for qualname in act_qualifiers:
            self.assertEqualCIMQualifier(
                act_qualifiers[qualname], exp_qualifiers[qualname],
                "%s, qualifier %s" % (context, qualname))

    def assertEqualCIMProperty(self, act_property, exp_property, context):
        """
        Verify that two CIMProperty declaration objects are equal.
        """

        self.assertEqual(
            act_property.name, exp_property.name,
            "%s (dict key), name attribute: %s (expected: %s)" %
            (context,
             act_property.name, exp_property.name))
        self.assertEqual(
            act_property.value, exp_property.value,
            "%s, value attribute: %s (expected: %s)" %
            (context,
             act_property.value, exp_property.value))
        self.assertEqual(
            act_property.type, exp_property.type,
            "%s, type attribute: %s (expected: %s)" %
            (context,
             act_property.type, exp_property.type))
        self.assertEqual(
            act_property.reference_class, exp_property.reference_class,
            "%s, reference_class attribute: %s (expected: %s)" %
            (context,
             act_property.reference_class, exp_property.reference_class))
        self.assertEqual(
            act_property.embedded_object, exp_property.embedded_object,
            "%s, embedded_object attribute: %s (expected: %s)" %
            (context,
             act_property.embedded_object, exp_property.embedded_object))
        self.assertEqual(
            act_property.is_array, exp_property.is_array,
            "%s, is_array attribute: %s (expected: %s)" %
            (context,
             act_property.is_array, exp_property.is_array))
        self.assertEqual(
            act_property.array_size, exp_property.array_size,
            "%s, array_size attribute: %s (expected: %s)" %
            (context,
             act_property.array_size, exp_property.array_size))
        self.assertEqual(
            act_property.class_origin, exp_property.class_origin,
            "%s, class_origin attribute: %s (expected: %s)" %
            (context,
             act_property.class_origin, exp_property.class_origin))
        self.assertEqual(
            act_property.propagated, exp_property.propagated,
            "%s, propagated attribute: %s (expected: %s)" %
            (context,
             act_property.propagated, exp_property.propagated))
        self.assertEqualQualifiers(
            act_property.qualifiers, exp_property.qualifiers,
            context)

    def assertEqualCIMMethod(self, act_method, exp_method, context):
        """
        Verify that two CIMMethod objects are equal.
        """

        self.assertEqual(
            act_method.name, exp_method.name,
            "%s (dict key), name attribute: %s (expected: %s)" %
            (context,
             act_method.name, exp_method.name))
        self.assertEqual(
            act_method.return_type, exp_method.return_type,
            "%s, return_type attribute: %s (expected: %s)" %
            (context,
             act_method.return_type, exp_method.return_type))
        self.assertEqual(
            act_method.class_origin, exp_method.class_origin,
            "%s, class_origin attribute: %s (expected: %s)" %
            (context,
             act_method.class_origin, exp_method.class_origin))
        self.assertEqual(
            act_method.propagated, exp_method.propagated,
            "%s, propagated attribute: %s methodmethod (expected: %s)" %
            (context,
             act_method.propagated, exp_method.propagated))
        self.assertEqualParameters(
            act_method.parameters, exp_method.parameters,
            context)
        self.assertEqualQualifiers(
            act_method.qualifiers, exp_method.qualifiers,
            context)

    def assertEqualCIMParameter(self, act_parameter, exp_parameter, context):
        """
        Verify that two CIMParameter objects are equal.
        """

        self.assertEqual(
            act_parameter.name, exp_parameter.name,
            "%s (dict key), name attribute: %s (expected: %s)" %
            (context,
             act_parameter.name, exp_parameter.name))
        self.assertEqual(
            act_parameter.type, exp_parameter.type,
            "%s, type attribute: %s (expected: %s)" %
            (context,
             act_parameter.type, exp_parameter.type))
        self.assertEqual(
            act_parameter.reference_class, exp_parameter.reference_class,
            "%s, reference_class attribute: %s (expected: %s)" %
            (context,
             act_parameter.reference_class, exp_parameter.reference_class))
        self.assertEqual(
            act_parameter.is_array, exp_parameter.is_array,
            "%s, is_array attribute: %s (expected: %s)" %
            (context,
             act_parameter.is_array, exp_parameter.is_array))
        self.assertEqual(
            act_parameter.array_size, exp_parameter.array_size,
            "%s, array_size attribute: %s (expected: %s)" %
            (context,
             act_parameter.array_size, exp_parameter.array_size))
        self.assertEqualQualifiers(
            act_parameter.qualifiers, exp_parameter.qualifiers,
            context)

    def assertEqualCIMQualifier(self, act_qualifier, exp_qualifier, context):
        """
        Verify that two CIMQualifier objects are equal.
        """

        self.assertEqual(
            act_qualifier.name, exp_qualifier.name,
            "%s (dict key), name attribute: %s (expected: %s)" %
            (context,
             act_qualifier.name, exp_qualifier.name))
        self.assertEqual(
            act_qualifier.value, exp_qualifier.value,
            "%s, value attribute: %s (expected: %s)" %
            (context,
             act_qualifier.value, exp_qualifier.value))
        self.assertEqual(
            act_qualifier.type, exp_qualifier.type,
            "%s, type attribute: %s (expected: %s)" %
            (context,
             act_qualifier.type, exp_qualifier.type))
        self.assertEqual(
            act_qualifier.propagated, exp_qualifier.propagated,
            "%s, propagated attribute: %s (expected: %s)" %
            (context,
             act_qualifier.propagated, exp_qualifier.propagated))
        self.assertEqual(
            act_qualifier.overridable, exp_qualifier.overridable,
            "%s, overridable attribute: %s (expected: %s)" %
            (context,
             act_qualifier.overridable, exp_qualifier.overridable))
        self.assertEqual(
            act_qualifier.tosubclass, exp_qualifier.tosubclass,
            "%s, tosubclass attribute: %s (expected: %s)" %
            (context,
             act_qualifier.tosubclass, exp_qualifier.tosubclass))
        self.assertEqual(
            act_qualifier.toinstance, exp_qualifier.toinstance,
            "%s, toinstance attribute: %s (expected: %s)" %
            (context,
             act_qualifier.toinstance, exp_qualifier.toinstance))
        self.assertEqual(
            act_qualifier.translatable, exp_qualifier.translatable,
            "%s, translatable attribute: %s (expected: %s)" %
            (context,
             act_qualifier.translatable, exp_qualifier.translatable))

    def assertEqualKeys(self, act_dict, exp_dict, context):
        """
        Verify that two dicts have the same set of keys.
        """
        act_keys = set(act_dict.keys())
        exp_keys = set(exp_dict.keys())
        if act_keys != exp_keys:
            missing_keys = exp_keys - act_keys
            added_keys = act_keys - exp_keys
            msg = "Different sets of %s" % context
            if missing_keys:
                msg += ", missing: %s" % list(missing_keys)
            if added_keys:
                msg += ", added: %s" % list(added_keys)
            self.fail(msg)



#
# (C) Copyright 2003,2004 Hewlett-Packard Development Company, L.P.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Martin Pool <mbp@hp.com>
# Author: Ross Peoples <ross.peoples@gmail.com>
#

"""
tupletree - Convert XML DOM objects to and from tuple trees.

DOM is the standard in-memory representation of XML documents, but it
is very cumbersome for some types of processing where XML encodes
object structures rather than text documents.  Direct mapping to Python
classes may not be a good match either.

tupletrees may be created from an in-memory DOM using
dom_to_tupletree(), or from a string using xml_to_tupletree().

Since the Python XML libraries deal mostly with Unicode strings they
are also returned here.  If plain Strings are passed in they will be
converted by xmldom.

Each node of the tuple tree is a Python 4-tuple, corresponding to an
XML Element (i.e. <tag>):

  (NAME, ATTRS, CONTENTS, None)

The NAME is the name of the element.

The ATTRS are a name-value hash of element attributes.

The CONTENTS is a list of child elements.

The fourth element is reserved.
"""

# NOTE: The original dom based parsers have been replaced with a sax parser
#       in Nov. 2016. The dom tupletree code is in the file
#       testsuite/test_tupletree.py

from __future__ import absolute_import

import xml.dom.minidom
import xml.sax
import re
import sys
import six

__all__ = []


class CIMContentHandler(xml.sax.ContentHandler):
    """SAX handler for CIM XML.

    Similar to dom_to_tupletree, the handler creates a tree of tuples,
    where the elements are the XML element name, the attributes, and the
    children.

    The handler pushes and pops the list of elements, building the child
    list as it goes.

    The end result is that the root node is left in the list and available
    as the root attribute of the object.
    """

    def __init__(self):
        xml.sax.ContentHandler.__init__(self)
        self.root = None
        self.elements = []
        self.element = []

    def startDocument(self):
        assert self.elements == []

    def endDocument(self):
        assert self.elements == []
        self.root = self.element

    def startElement(self, name, attrs):
        if self.element:
            self.elements.append(self.element)
        # Avoid dict comprehension to support python2.6.
        attr_dict = {}
        for k, v in attrs.items():
            attr_dict[k] = v
        element = (name, attr_dict, list(), None)
        if self.element:
            self.element[2].append(element)
        self.element = element

    def endElement(self, name):
        if self.elements:
            self.element = self.elements.pop()

    def characters(self, content):
        if self.element[2]:
            try:
                re.match(r'\s+', self.element[2][-1])
            except TypeError:
                pass
            else:
                ws = self.element[2].pop()
                content = '%s%s' % (ws, content)
        self.element[2].append(content)


def xml_to_tupletree_sax(xml_string):
    """
    Parse an XML string into tupletree with SAX parser.

    Parses the string using the class CIMContentHandler and
    returns the root element. As a SAX parser it uses minimal
    memory.

    This is a replacement for the previous parser (xml_to_tuple)
    which used the dom parser.

    Parameters:

      xml_string (:term:`string`) or (:term:`byte string`): A string
      or bytes string containing the XML to be parsed.

    Returns:
      (:class: `CIMContentHandler`) : The root tuple from the
      parse:

    Raises:
        TypeError, ParseError, SAXParseException,  or ExpatError.
    """
    handler = CIMContentHandler()

    # The following is required because the SAX parser in these Python
    # versions does not accept unicode strings, raising:
    # SAXParseException: "<unknown>:1:1: not well-formed (invalid token)"
    if sys.version_info[0:2] in ((2, 6), (3, 4)):
        if isinstance(xml_string, six.text_type):
            xml_string = xml_string.encode("utf-8")

    xml.sax.parseString(xml_string, handler, None)
    return handler.root
